"""
string_tokenizer.py

A utility module for string tokenization operations.
Provides methods to split strings into tokens and manipulate string boundaries.

Copyright (c) 2025 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""


class StringTokenizer:
    """
    Utility class for string tokenization operations.

    This class provides methods to:
    - Split strings into tokens based on delimiters
    - Process multiple strings into token lists
    - Remove matching characters from string boundaries
    """

    @staticmethod
    def parse(
        content: str,
        delimiter: str,
        *,
        strip: bool = True
    ) -> list[str]:
        """
        Split a string into tokens based on a delimiter.

        Args:
            content (str): The input string to tokenize
            delimiter (str): The character(s) to split the string on
            strip (bool, optional): Whether to strip whitespace from tokens. Defaults to True.

        Returns:
            list[str]: List of tokens, with empty tokens removed if strip is True

        Raises:
            ValueError: If delimiter is empty or None

        Example:
            >>> StringTokenizer.parse("a,b,c", ",")
            ['a', 'b', 'c']
        """
        if not delimiter:
            raise ValueError("Delimiter cannot be empty or None")

        if content is None:
            return []

        if strip and not content.strip():
            return []

        result: list[str] = content.split(delimiter)
        if strip:
            result = [token.strip() for token in result if token.strip()]
        return result

    @classmethod
    def parses(
        cls,
        content: list[str],
        delimiter: str,
        *,
        strip: bool = True
    ) -> list[list[str]]:
        """
        Process multiple strings into lists of tokens.

        Args:
            content (list[str]): List of strings to tokenize
            delimiter (str): The character(s) to split each string on
            strip (bool, optional): Whether to strip whitespace from tokens. Defaults to True.

        Returns:
            list[list[str]]: List of token lists, one for each input string

        Example:
            >>> StringTokenizer.parses(["a,b", "c,d"], ",")
            [['a', 'b'], ['c', 'd']]
        """
        return [cls.parse(text, delimiter, strip=strip) for text in content]

    @staticmethod
    def remove_bookends(
        content: str,
        bookend: str,
        *,
        strip: bool = True
    ) -> str:
        """
        Remove matching characters from both ends of a string.

        Args:
            content (str): The input string to process
            bookend (str): The character(s) to remove from both ends
            strip (bool, optional): Whether to strip whitespace first. Defaults to True.

        Returns:
            str: The string with matching bookends removed

        Example:
            >>> StringTokenizer.remove_bookends("'hello'", "'")
            'hello'
        """
        value: str = content.strip() if strip else content
        if (
            value.startswith(bookend)
            and value.endswith(bookend)
            and len(value) > 2 * len(bookend) - 1
        ):
            return value[len(bookend) : -len(bookend)]
        return value



"""
Unit tests for TypedTabularDataModel.

Copyright (c) 2025, Jim Schilling

Please keep the copyright notice in this file and in the source code files.

This module is licensed under the MIT License.
"""

import unittest
from datetime import date, datetime, time

from splurge_tools.type_helper import DataType
from splurge_tools.typed_tabular_data_model import TypeConfig, TypedTabularDataModel


class TestTypedTabularDataModel(unittest.TestCase):
    """Test cases for TypedTabularDataModel."""

    def setUp(self):
        """Set up test data."""
        self.test_data = [
            ["name", "age", "is_active", "score", "birth_date", "created_at", "login_time"],
            ["John", "25", "true", "95.5", "1998-01-01", "2024-01-01T12:00:00", "14:30:45"],
            ["Jane", "", "false", "", "1993-05-15", "2024-01-02T13:00:00", "2:30 PM"],
            ["Bob", "none", "none", "none", "none", "none", "none"],
            ["Alice", "30", "true", "88.0", "1995-12-31", "2024-01-03T14:00:00", ""],
        ]

        # Custom type configurations
        self.custom_configs = {
            DataType.BOOLEAN: TypeConfig(empty_default=False, none_default=True),
            DataType.INTEGER: TypeConfig(empty_default=-1, none_default=0),
            DataType.FLOAT: TypeConfig(empty_default=0.0, none_default=0.0),
            DataType.DATE: TypeConfig(
                empty_default=date(1900, 1, 1), none_default=None
            ),
            DataType.DATETIME: TypeConfig(
                empty_default=datetime(1900, 1, 1), none_default=None
            ),
            DataType.TIME: TypeConfig(
                empty_default=time(0, 0, 0), none_default=None
            ),
        }

        # Create models with default and custom configurations
        self.default_model = TypedTabularDataModel(self.test_data)
        self.custom_model = TypedTabularDataModel(
            self.test_data, type_configs=self.custom_configs
        )

    def test_column_types(self):
        """Test that column types are correctly inferred."""
        self.assertEqual(self.default_model.column_type("name"), DataType.STRING)
        self.assertEqual(self.default_model.column_type("age"), DataType.INTEGER)
        self.assertEqual(self.default_model.column_type("is_active"), DataType.BOOLEAN)
        self.assertEqual(self.default_model.column_type("score"), DataType.FLOAT)
        self.assertEqual(self.default_model.column_type("birth_date"), DataType.DATE)
        self.assertEqual(
            self.default_model.column_type("created_at"), DataType.DATETIME
        )
        self.assertEqual(
            self.default_model.column_type("login_time"), DataType.TIME
        )

    def test_default_conversions(self):
        """Test type conversions with default configurations."""
        # Test normal values
        self.assertEqual(self.default_model.cell_value("name", 0), "John")
        self.assertEqual(self.default_model.cell_value("age", 0), 25)
        self.assertEqual(self.default_model.cell_value("is_active", 0), True)
        self.assertEqual(self.default_model.cell_value("score", 0), 95.5)
        self.assertEqual(
            self.default_model.cell_value("birth_date", 0), date(1998, 1, 1)
        )
        self.assertEqual(
            self.default_model.cell_value("created_at", 0), datetime(2024, 1, 1, 12, 0)
        )
        self.assertEqual(
            self.default_model.cell_value("login_time", 0), time(14, 30, 45)
        )

        # Test empty values
        self.assertEqual(
            self.default_model.cell_value("age", 1), 0
        )  # empty_default for INTEGER
        self.assertEqual(
            self.default_model.cell_value("score", 1), 0.0
        )  # empty_default for FLOAT
        self.assertEqual(
            self.default_model.cell_value("is_active", 1), False
        )  # empty_default for BOOLEAN
        self.assertEqual(
            self.default_model.cell_value("login_time", 3), None
        )  # none_default for TIME
        self.assertEqual(
            self.default_model.cell_value("login_time", 1), time(14, 30)
        )  # 12-hour format
        self.assertEqual(
            self.default_model.cell_value("login_time", 2), None
        )  # none_default for TIME
        self.assertEqual(
            self.default_model.cell_value("login_time", 4-1), None
        )  # empty string, default is None

        # Test none-like values
        self.assertEqual(
            self.default_model.cell_value("age", 2), 0
        )  # none_default for INTEGER
        self.assertEqual(
            self.default_model.cell_value("score", 2), 0.0
        )  # none_default for FLOAT
        self.assertEqual(
            self.default_model.cell_value("is_active", 2), False
        )  # none_default for BOOLEAN
        self.assertEqual(
            self.default_model.cell_value("login_time", 2), None
        )  # none_default for TIME

    def test_custom_conversions(self):
        """Test type conversions with custom configurations."""
        # Test normal values (should be same as default)
        self.assertEqual(self.custom_model.cell_value("name", 0), "John")
        self.assertEqual(self.custom_model.cell_value("age", 0), 25)
        self.assertEqual(self.custom_model.cell_value("is_active", 0), True)
        self.assertEqual(self.custom_model.cell_value("score", 0), 95.5)
        self.assertEqual(
            self.custom_model.cell_value("birth_date", 0), date(1998, 1, 1)
        )
        self.assertEqual(
            self.custom_model.cell_value("created_at", 0), datetime(2024, 1, 1, 12, 0)
        )
        self.assertEqual(
            self.custom_model.cell_value("login_time", 0), time(14, 30, 45)
        )

        # Test empty values with custom defaults
        self.assertEqual(
            self.custom_model.cell_value("age", 1), -1
        )  # custom empty_default for INTEGER
        self.assertEqual(
            self.custom_model.cell_value("score", 1), 0.0
        )  # custom empty_default for FLOAT
        self.assertEqual(
            self.custom_model.cell_value("is_active", 1), False
        )  # custom empty_default for BOOLEAN
        self.assertEqual(
            self.custom_model.cell_value("birth_date", 1), date(1993, 5, 15)
        )  # actual date value
        self.assertEqual(
            self.custom_model.cell_value("login_time", 3), time(0, 0, 0)
        )  # custom empty_default for TIME
        self.assertEqual(
            self.custom_model.cell_value("login_time", 1), time(14, 30)
        )  # 12-hour format
        self.assertEqual(
            self.custom_model.cell_value("login_time", 2), None
        )  # custom none_default for TIME
        self.assertEqual(
            self.custom_model.cell_value("login_time", 4-1), time(0, 0, 0)
        )  # custom empty_default for TIME

        # Test none-like values with custom defaults
        self.assertEqual(
            self.custom_model.cell_value("age", 2), 0
        )  # custom none_default for INTEGER
        self.assertEqual(
            self.custom_model.cell_value("score", 2), 0.0
        )  # custom none_default for FLOAT
        self.assertEqual(
            self.custom_model.cell_value("is_active", 2), True
        )  # custom none_default for BOOLEAN
        self.assertEqual(
            self.custom_model.cell_value("birth_date", 2), None
        )  # custom none_default for DATE
        self.assertEqual(
            self.custom_model.cell_value("login_time", 2), None
        )  # custom none_default for TIME

    def test_column_values(self):
        """Test getting all values for a column."""
        # Test with default configuration
        ages = self.default_model.column_values("age")
        self.assertEqual(ages, [25, 0, 0, 30])

        scores = self.default_model.column_values("score")
        self.assertEqual(scores, [95.5, 0.0, 0.0, 88.0])

        login_times = self.default_model.column_values("login_time")
        self.assertEqual(login_times, [time(14, 30, 45), time(14, 30), None, None])

        # Test with custom configuration
        ages = self.custom_model.column_values("age")
        self.assertEqual(ages, [25, -1, 0, 30])

        scores = self.custom_model.column_values("score")
        self.assertEqual(scores, [95.5, 0.0, 0.0, 88.0])

        login_times = self.custom_model.column_values("login_time")
        self.assertEqual(login_times, [time(14, 30, 45), time(14, 30), None, time(0, 0, 0)])

    def test_row_access(self):
        """Test accessing rows in different formats."""
        # Test dictionary access
        row = self.default_model.row(0)
        self.assertEqual(row["name"], "John")
        self.assertEqual(row["age"], 25)
        self.assertEqual(row["is_active"], True)
        self.assertEqual(row["score"], 95.5)
        self.assertEqual(row["birth_date"], date(1998, 1, 1))
        self.assertEqual(row["created_at"], datetime(2024, 1, 1, 12, 0))
        self.assertEqual(row["login_time"], time(14, 30, 45))

        # Test list access
        row_list = self.default_model.row_as_list(0)
        self.assertEqual(row_list[0], "John")
        self.assertEqual(row_list[1], 25)
        self.assertEqual(row_list[2], True)
        self.assertEqual(row_list[3], 95.5)
        self.assertEqual(row_list[4], date(1998, 1, 1))
        self.assertEqual(row_list[5], datetime(2024, 1, 1, 12, 0))
        self.assertEqual(row_list[6], time(14, 30, 45))

        # Test tuple access
        row_tuple = self.default_model.row_as_tuple(0)
        self.assertEqual(row_tuple[0], "John")
        self.assertEqual(row_tuple[1], 25)
        self.assertEqual(row_tuple[2], True)
        self.assertEqual(row_tuple[3], 95.5)
        self.assertEqual(row_tuple[4], date(1998, 1, 1))
        self.assertEqual(row_tuple[5], datetime(2024, 1, 1, 12, 0))
        self.assertEqual(row_tuple[6], time(14, 30, 45))

    def test_iterators(self):
        """Test row iterators."""
        # Test dictionary iterator
        rows = list(self.default_model.iter_rows())
        self.assertEqual(len(rows), 4)  # 4 data rows
        self.assertEqual(rows[0]["name"], "John")
        self.assertEqual(rows[0]["age"], 25)
        self.assertEqual(rows[0]["login_time"], time(14, 30, 45))

        # Test tuple iterator
        rows = list(self.default_model.iter_rows_as_tuples())
        self.assertEqual(len(rows), 4)  # 4 data rows
        self.assertEqual(rows[0][0], "John")
        self.assertEqual(rows[0][1], 25)
        self.assertEqual(rows[0][6], time(14, 30, 45))

    def test_invalid_column(self):
        """Test handling of invalid column names."""
        with self.assertRaises(ValueError):
            self.default_model.column_values("invalid_column")

        with self.assertRaises(ValueError):
            self.default_model.cell_value("invalid_column", 0)

    def test_invalid_row_index(self):
        """Test handling of invalid row indices."""
        with self.assertRaises(ValueError):
            self.default_model.cell_value("name", -1)

        with self.assertRaises(ValueError):
            self.default_model.cell_value("name", 10)

        with self.assertRaises(ValueError):
            self.default_model.row(-1)

        with self.assertRaises(ValueError):
            self.default_model.row(10)

    def test_mixed_type_handling(self):
        """Test handling of MIXED type values."""
        # Create test data with a MIXED type column
        mixed_data = [
            ["id", "mixed_col"],
            ["1", "123"],  # integer-like
            ["2", "abc"],  # string
            ["3", ""],  # empty
            ["4", "none"],  # none-like
        ]

        # Create model with custom config for MIXED type
        mixed_configs = {
            DataType.MIXED: TypeConfig(empty_default="EMPTY", none_default="NONE")
        }
        mixed_model = TypedTabularDataModel(mixed_data, type_configs=mixed_configs)

        # Test that MIXED type values are preserved
        self.assertEqual(
            mixed_model.cell_value("mixed_col", 0), "123"
        )  # integer-like preserved as string
        self.assertEqual(
            mixed_model.cell_value("mixed_col", 1), "abc"
        )  # string preserved
        self.assertEqual(
            mixed_model.cell_value("mixed_col", 2), ""
        )  # empty value preserved
        self.assertEqual(
            mixed_model.cell_value("mixed_col", 3), "NONE"
        )  # none-like uses none_default

        # Test column values
        mixed_values = mixed_model.column_values("mixed_col")
        self.assertEqual(mixed_values, ["123", "abc", "", "NONE"])


if __name__ == "__main__":
    unittest.main()



"""
test_string_tokenizer.py

Unit tests for the StringTokenizer class.
"""

import unittest

from splurge_tools.string_tokenizer import StringTokenizer


class TestStringTokenizer(unittest.TestCase):
    """Test cases for the StringTokenizer class."""

    def test_parse_basic(self):
        """Test basic string parsing functionality."""
        result = StringTokenizer.parse("a,b,c", ",")
        self.assertEqual(result, ["a", "b", "c"])

    def test_parse_with_spaces(self):
        """Test parsing with whitespace handling."""
        result = StringTokenizer.parse("a, b , c", ",")
        self.assertEqual(result, ["a", "b", "c"])

    def test_parse_empty_tokens(self):
        """Test parsing with empty tokens."""
        result = StringTokenizer.parse("a,,c", ",")
        self.assertEqual(result, ["a", "c"])

    def test_parse_no_strip(self):
        """Test parsing without stripping whitespace."""
        result = StringTokenizer.parse("a, b , c", ",", strip=False)
        self.assertEqual(result, ["a", " b ", " c"])

    def test_parses_basic(self):
        """Test parsing multiple strings."""
        result = StringTokenizer.parses(["a,b", "c,d"], ",")
        self.assertEqual(result, [["a", "b"], ["c", "d"]])

    def test_parses_with_spaces(self):
        """Test parsing multiple strings with whitespace."""
        result = StringTokenizer.parses(["a, b", "c, d"], ",")
        self.assertEqual(result, [["a", "b"], ["c", "d"]])

    def test_remove_bookends_basic(self):
        """Test basic bookend removal."""
        result = StringTokenizer.remove_bookends("'hello'", "'")
        self.assertEqual(result, "hello")

    def test_remove_bookends_no_match(self):
        """Test bookend removal when no match."""
        result = StringTokenizer.remove_bookends("hello", "'")
        self.assertEqual(result, "hello")

    def test_remove_bookends_single_char(self):
        """Test bookend removal with single character."""
        result = StringTokenizer.remove_bookends("'a'", "'")
        self.assertEqual(result, "a")

    def test_remove_bookends_with_spaces(self):
        """Test bookend removal with surrounding spaces."""
        result = StringTokenizer.remove_bookends("  'hello'  ", "'")
        self.assertEqual(result, "hello")

    def test_remove_bookends_no_strip(self):
        """Test bookend removal without stripping."""
        result = StringTokenizer.remove_bookends("  'hello'  ", "'", strip=False)
        self.assertEqual(result, "  'hello'  ")


if __name__ == "__main__":
    unittest.main()



"""Unit tests for text_normalizer.py"""

import unittest

from splurge_tools.text_normalizer import TextNormalizer


class TestTextNormalizer(unittest.TestCase):
    def test_remove_accents(self):
        self.assertEqual(TextNormalizer.remove_accents("café"), "cafe")
        self.assertEqual(TextNormalizer.remove_accents("résumé"), "resume")
        self.assertEqual(TextNormalizer.remove_accents(""), "")
        self.assertEqual(TextNormalizer.remove_accents(None), "")

    def test_normalize_whitespace(self):
        # Test without preserving newlines
        self.assertEqual(
            TextNormalizer.normalize_whitespace("hello   world"), "hello world"
        )
        self.assertEqual(
            TextNormalizer.normalize_whitespace("hello\n\nworld"), "hello world"
        )
        self.assertEqual(
            TextNormalizer.normalize_whitespace("  hello  world  "), "hello world"
        )

        # Test with preserving newlines
        self.assertEqual(
            TextNormalizer.normalize_whitespace(
                "hello\n\nworld", preserve_newlines=True
            ),
            "hello\n\nworld",
        )
        self.assertEqual(
            TextNormalizer.normalize_whitespace(
                "hello   world\n\n  today", preserve_newlines=True
            ),
            "hello world\n\ntoday",
        )

    def test_remove_special_chars(self):
        self.assertEqual(
            TextNormalizer.remove_special_chars("hello@world!"), "helloworld"
        )
        self.assertEqual(
            TextNormalizer.remove_special_chars("hello@world!", keep_chars="@"),
            "hello@world",
        )
        self.assertEqual(TextNormalizer.remove_special_chars(""), "")
        self.assertEqual(TextNormalizer.remove_special_chars(None), "")

    def test_normalize_line_endings(self):
        self.assertEqual(
            TextNormalizer.normalize_line_endings("hello\r\nworld"), "hello\nworld"
        )
        self.assertEqual(
            TextNormalizer.normalize_line_endings("hello\rworld"), "hello\nworld"
        )
        self.assertEqual(
            TextNormalizer.normalize_line_endings("hello\nworld", line_ending="\r\n"),
            "hello\r\nworld",
        )

    def test_to_ascii(self):
        self.assertEqual(TextNormalizer.to_ascii("café"), "cafe")
        self.assertEqual(TextNormalizer.to_ascii("résumé"), "resume")
        self.assertEqual(TextNormalizer.to_ascii("café", replacement="x"), "cafe")

    def test_remove_control_chars(self):
        self.assertEqual(
            TextNormalizer.remove_control_chars("hello\x00world"), "helloworld"
        )
        self.assertEqual(
            TextNormalizer.remove_control_chars("hello\x1fworld"), "helloworld"
        )

    def test_normalize_quotes(self):
        # Test with double quotes
        self.assertEqual(
            TextNormalizer.normalize_quotes('hello "world"'), 'hello "world"'
        )
        # Test with single quotes
        self.assertEqual(
            TextNormalizer.normalize_quotes("hello 'world'"), 'hello "world"'
        )
        # Test with mixed quotes
        self.assertEqual(
            TextNormalizer.normalize_quotes('hello "world\'s"'), 'hello "world\'s"'
        )
        # Test with custom quote character
        self.assertEqual(
            TextNormalizer.normalize_quotes("hello 'world'", quote_char="'"),
            "hello 'world'",
        )
        # Test with empty string
        self.assertEqual(TextNormalizer.normalize_quotes(""), "")
        # Test with None
        self.assertEqual(TextNormalizer.normalize_quotes(None), "")
        # Test with apostrophes
        self.assertEqual(
            TextNormalizer.normalize_quotes("it's a 'test'"), 'it\'s a "test"'
        )

    def test_normalize_dashes(self):
        self.assertEqual(TextNormalizer.normalize_dashes("hello–world"), "hello-world")
        self.assertEqual(TextNormalizer.normalize_dashes("hello—world"), "hello-world")
        # Test with existing dash
        self.assertEqual(TextNormalizer.normalize_dashes("hello-world"), "hello-world")
        # Test with custom dash character
        self.assertEqual(
            TextNormalizer.normalize_dashes("hello–world", dash_char="_"), "hello_world"
        )

    def test_normalize_spaces(self):
        self.assertEqual(
            TextNormalizer.normalize_spaces("hello\u00a0world"), "hello world"
        )
        self.assertEqual(TextNormalizer.normalize_spaces("hello  world"), "hello world")

    def test_normalize_case(self):
        self.assertEqual(
            TextNormalizer.normalize_case("Hello World", case="lower"), "hello world"
        )
        self.assertEqual(
            TextNormalizer.normalize_case("hello world", case="upper"), "HELLO WORLD"
        )
        self.assertEqual(
            TextNormalizer.normalize_case("hello world", case="title"), "Hello World"
        )
        self.assertEqual(
            TextNormalizer.normalize_case("hello world", case="sentence"), "Hello world"
        )

    def test_remove_duplicate_chars(self):
        # Test default behavior (spaces and dashes)
        self.assertEqual(
            TextNormalizer.remove_duplicate_chars("hello   world"), "hello world"
        )
        self.assertEqual(
            TextNormalizer.remove_duplicate_chars("hello--world"), "hello-world"
        )

        # Test with periods
        self.assertEqual(
            TextNormalizer.remove_duplicate_chars("hello...world"), "hello.world"
        )

        # Test with custom characters
        self.assertEqual(
            TextNormalizer.remove_duplicate_chars("hello...world---today", chars=".-"),
            "hello.world-today",
        )

        # Test with no duplicates
        self.assertEqual(
            TextNormalizer.remove_duplicate_chars("hello world"), "hello world"
        )


if __name__ == "__main__":
    unittest.main()



"""
A utility module for text normalization operations.

Copyright (c) 2025 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""

import re
import unicodedata
from functools import wraps
from typing import Any, Callable, Pattern

from splurge_tools.case_helper import CaseHelper


def handle_empty_value(
    func: Callable[[str, Any], str]
) -> Callable[[str, Any], str]:
    """Decorator to handle empty value checks for normalization methods."""
    @wraps(func)
    def wrapper(value: str, *args: Any, **kwargs: Any) -> str:
        if value is None or not value:
            return ""
        return func(value, *args, **kwargs)
    return wrapper


class TextNormalizer:
    """
    A utility class for text normalization operations.

    This class provides methods to:
    - Remove diacritics (accents)
    - Normalize whitespace
    - Remove special characters
    - Normalize line endings
    - Convert to ASCII
    - Remove control characters
    - Normalize quotes
    - Normalize dashes
    - Normalize spaces
    - Normalize case

    All methods support an optional normalize parameter (default: True) that:
    - Handles empty values
    - Preserves original string if no changes needed
    """

    _WHITESPACE_PATTERN: Pattern = re.compile(r"\s+")
    _CONTROL_CHARS_PATTERN: Pattern = re.compile(r"[\x00-\x1f\x7f-\x9f]")
    _SPECIAL_CHARS_PATTERN: Pattern = re.compile(r"[^\w\s-]")

    @classmethod
    @handle_empty_value
    def remove_accents(
        cls,
        value: str
    ) -> str:
        """
        Remove accents from text.

        Args:
            value: Input string to normalize

        Returns:
            String with diacritical marks removed

        Example:
            "café" -> "cafe"
            "résumé" -> "resume"
        """
        if value is None:
            return ""
        return "".join(
            c
            for c in unicodedata.normalize("NFKD", value)
            if not unicodedata.combining(c)
        )

    @classmethod
    @handle_empty_value
    def normalize_whitespace(
        cls,
        value: str,
        preserve_newlines: bool = False
    ) -> str:
        """
        Normalize whitespace in text.

        Args:
            value: Input string to normalize
            preserve_newlines: Whether to preserve newline characters

        Returns:
            String with normalized whitespace

        Example:
            "hello   world" -> "hello world"
            "hello\n\nworld" -> "hello world" (if preserve_newlines=False)
        """
        if preserve_newlines:
            value = re.sub(r"[^\S\n]+", " ", value)
            value = re.sub(r"\r\n|\r", "\n", value)
            value = re.sub(r"\n\s*\n", "\n\n", value)
            value = re.sub(r" +(\n)", r"\1", value)
            value = re.sub(r"(\n) +", r"\1", value)
        else:
            value = cls._WHITESPACE_PATTERN.sub(" ", value)
        return value.strip()

    @classmethod
    @handle_empty_value
    def remove_special_chars(
        cls,
        value: str,
        keep_chars: str = ""
    ) -> str:
        """
        Remove special characters from text.

        Args:
            value: Input string to normalize
            keep_chars: Additional characters to preserve

        Returns:
            String with special characters removed

        Example:
            "hello@world!" -> "helloworld"
            "hello@world!" (keep_chars="@") -> "hello@world"
        """
        if value is None:
            return ""
        pattern: str = f"[^\\w\\s{re.escape(keep_chars)}]"
        return re.sub(pattern, "", value)

    @classmethod
    @handle_empty_value
    def normalize_line_endings(
        cls,
        value: str,
        line_ending: str = "\n"
    ) -> str:
        """
        Normalize line endings in text.

        Args:
            value: Input string to normalize
            line_ending: Desired line ending character

        Returns:
            String with normalized line endings

        Example:
            "hello\r\nworld" -> "hello\nworld"
        """
        return re.sub(r"\r\n|\r|\n", line_ending, value)

    @classmethod
    @handle_empty_value
    def to_ascii(
        cls,
        value: str,
        *,
        replacement: str = ""
    ) -> str:
        """
        Convert text to ASCII, replacing non-ASCII characters.

        Args:
            value: Input string to normalize
            replacement: Character to use for non-ASCII characters

        Returns:
            ASCII string

        Example:
            "café" -> "cafe"
            "résumé" -> "resume"
        """
        value = cls.remove_accents(value)
        return (
            value.encode("ascii", "replace").decode("ascii").replace("?", replacement)
        )

    @classmethod
    @handle_empty_value
    def remove_control_chars(
        cls,
        value: str
    ) -> str:
        """
        Remove control characters from text.

        Args:
            value: Input string to normalize

        Returns:
            String with control characters removed

        Example:
            "hello\x00world" -> "helloworld"
        """
        return cls._CONTROL_CHARS_PATTERN.sub("", value)

    @classmethod
    @handle_empty_value
    def normalize_quotes(
        cls,
        value: str,
        *,
        quote_char: str = '"'
    ) -> str:
        """
        Normalize quote characters in text.

        Args:
            value: Input string to normalize
            quote_char: Desired quote character

        Returns:
            String with normalized quotes

        Example:
            'hello "world"' -> 'hello "world"'
            "hello 'world'" -> 'hello "world"'
            "hello 'world's" -> 'hello "world's"'
        """
        if value is None:
            return ""
        temp: str = re.sub(r"(\w)'(\w)", r"\1§APOS§\2", value)
        temp = temp.replace('"', quote_char).replace("'", quote_char)
        result: str = temp.replace("§APOS§", "'")
        return result

    @classmethod
    @handle_empty_value
    def normalize_dashes(
        cls,
        value: str,
        *,
        dash_char: str = "-"
    ) -> str:
        """
        Normalize dash characters in text.

        Args:
            value: Input string to normalize
            dash_char: Desired dash character

        Returns:
            String with normalized dashes

        Example:
            "hello–world" -> "hello-world"
            "hello—world" -> "hello-world"
        """
        return re.sub(r"[–—]", dash_char, value)

    @classmethod
    @handle_empty_value
    def normalize_spaces(
        cls,
        value: str
    ) -> str:
        """
        Normalize space characters in text.

        Args:
            value: Input string to normalize

        Returns:
            String with normalized spaces

        Example:
            "hello\u00a0world" -> "hello world"
        """
        return " ".join(value.split())

    @classmethod
    @handle_empty_value
    def normalize_case(
        cls,
        value: str,
        *,
        case: str = "lower"
    ) -> str:
        """
        Normalize text case.

        Args:
            value: Input string to normalize
            case: Desired case ('lower', 'upper', 'title', 'sentence')

        Returns:
            String with normalized case

        Example:
            "Hello World" (case='lower') -> "hello world"
            "hello world" (case='title') -> "Hello World"
        """
        case = case.lower()
        if case == "lower":
            return value.lower()
        if case == "upper":
            return value.upper()
        if case == "title":
            return value.title()
        if case == "sentence":
            return CaseHelper.to_sentence(value)
        return value

    @classmethod
    @handle_empty_value
    def remove_duplicate_chars(
        cls,
        value: str,
        *,
        chars: str = " -."
    ) -> str:
        """
        Remove embedded duplicate characters from text.

        Args:
            value: Input string to normalize
            chars: String of characters to deduplicate (default: space and dash)

        Returns:
            String with duplicate characters removed

        Example:
            "hello   world" -> "hello world"
            "hello--world" -> "hello-world"
            "hello...world" (chars='.') -> "hello.world"
        """
        result: str = value
        for char in chars:
            pattern: str = f"{re.escape(char)}{{2,}}"
            result = re.sub(pattern, char, result)
        return result



"""
type_helper module - Provides utilities for type checking, conversion and inference

This module offers a comprehensive set of tools for:
- Type inference and validation
- Data type conversion
- String parsing and validation
- Collection type checking
- Value profiling and analysis

Copyright (c) 2023 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""

import collections.abc
import re
import typing
from datetime import date, datetime, time
from enum import Enum
from typing import Any, Iterable, Union


class DataType(Enum):
    """
    Enumeration of supported data types for type inference and conversion.

    This enum defines the core data types that can be inferred and converted:
    - STRING: Text data
    - INTEGER: Whole numbers
    - FLOAT: Decimal numbers
    - BOOLEAN: True/False values
    - DATE: Calendar dates
    - TIME: Time values
    - DATETIME: Combined date and time
    - MIXED: Multiple types in collection
    - EMPTY: Empty values
    - NONE: Null/None values
    """

    STRING = "str"
    INTEGER = "int"
    FLOAT = "float"
    BOOLEAN = "bool"
    DATE = "date"
    TIME = "time"
    DATETIME = "datetime"
    MIXED = "mixed"
    EMPTY = "empty"
    NONE = "none"


class String:
    """
    Utility class for string type checking and conversion operations.

    This class provides static methods for:
    - Type validation (is_*_like methods)
    - Type conversion (to_* methods)
    - Type inference
    - String format validation
    """

    @staticmethod
    def is_bool_like(
        value: Union[str, bool, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a boolean.

        Args:
            value: Value to check (string or bool)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is bool or string 'true'/'false'

        Examples:
            >>> String.is_bool_like('true')  # True
            >>> String.is_bool_like('false') # True
            >>> String.is_bool_like('yes')   # False
        """
        if isinstance(value, bool):
            return True

        if isinstance(value, str):
            tmp_value: str = value.lower().strip() if trim else value.lower()
            return tmp_value in ["true", "false"]

        return False

    @staticmethod
    def is_none_like(
        value: Any,
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value represents None/null.

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is None or string 'none'/'null'

        Examples:
            >>> String.is_none_like(None)    # True
            >>> String.is_none_like('none')  # True
            >>> String.is_none_like('null')  # True
        """
        if value is None:
            return True

        if isinstance(value, str):
            tmp_value: str = value.strip().lower() if trim else value.lower()
            return tmp_value in ["none", "null"]

        return False

    @staticmethod
    def is_empty_like(
        value: Any,
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value is an empty string or contains only whitespace.

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is empty string or contains only whitespace

        Examples:
            >>> String.is_empty_like('')      # True
            >>> String.is_empty_like('   ')   # True
            >>> String.is_empty_like('abc')   # False
            >>> String.is_empty_like(None)    # False
        """
        if not isinstance(value, str):
            return False

        return not value.strip() if trim else not value

    @staticmethod
    def is_float_like(
        value: Union[str, float, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a float.

        Args:
            value: Value to check (string or float)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is float or string representing a float

        Examples:
            >>> String.is_float_like('1.23')  # True
            >>> String.is_float_like('-1.23') # True
            >>> String.is_float_like('1')     # False
        """
        if value is None:
            return False

        if isinstance(value, float):
            return True

        if not isinstance(value, str):
            return False

        return (
            re.match(r"""^[-+]?(\d+)?\.(\d+)?$""", value.strip() if trim else value)
            is not None
        )

    @staticmethod
    def is_int_like(
        value: Union[str, int, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as an integer.

        Args:
            value: Value to check (string or int)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is int or string representing an integer

        Examples:
            >>> String.is_int_like('123')   # True
            >>> String.is_int_like('-123')  # True
            >>> String.is_int_like('1.23')  # False
        """
        if value is None:
            return False

        if isinstance(value, int):
            return True

        if not isinstance(value, str):
            return False

        tmp_value: str = value.strip() if trim else value
        return re.match(r"""^[-+]?\d+$""", tmp_value) is not None

    @classmethod
    def is_numeric_like(
        cls,
        value: Union[str, float, int, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a number (int or float).

        Args:
            value: Value to check (string, float, or int)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is numeric or string representing a number

        Examples:
            >>> String.is_numeric_like('123')   # True
            >>> String.is_numeric_like('1.23')  # True
            >>> String.is_numeric_like('abc')   # False
        """
        if value is None:
            return False

        return cls.is_float_like(value, trim=trim) or cls.is_int_like(value, trim=trim)

    @classmethod
    def is_category_like(
        cls,
        value: Union[str, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value is non-numeric (categorical).

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is not numeric

        Examples:
            >>> String.is_category_like('abc')   # True
            >>> String.is_category_like('123')   # False
            >>> String.is_category_like('1.23')  # False
        """
        if value is None:
            return False

        return not cls.is_numeric_like(value, trim=trim)

    @staticmethod
    def _is_date_like(
        value: str
    ) -> bool:
        """
        Internal method to check if string matches common date formats.

        Args:
            value: String to check

        Returns:
            True if string matches a supported date format

        Note:
            Supports multiple date formats including:
            - YYYY-MM-DD
            - YYYY/MM/DD
            - YYYY.MM.DD
            - YYYYMMDD
            And their variations with different date component orders
        """
        patterns: list[str] = [
            "%Y-%m-%d",
            "%Y/%m/%d",
            "%Y.%m.%d",
            "%Y%m%d",
            "%Y-%d-%m",
            "%Y/%d/%m",
            "%Y.%d.%m",
            "%Y%d%m",
            "%m-%d-%Y",
            "%m/%d/%Y",
            "%m.%d.%Y",
            "%m%d%Y",
        ]

        for pattern in patterns:
            try:
                datetime.strptime(value, pattern)
                return True
            except ValueError:
                pass

        return False

    @staticmethod
    def _is_time_like(
        value: str
    ) -> bool:
        """
        Internal method to check if string matches common time formats.

        Args:
            value: String to check

        Returns:
            True if string matches a supported time format

        Note:
            Supports multiple time formats including:
            - HH:MM:SS
            - HH:MM:SS.microseconds
            - HH:MM
            - HHMMSS
            - HHMM
            And 12-hour format variations with AM/PM
        """
        patterns: list[str] = [
            "%H:%M:%S",
            "%H:%M:%S.%f",
            "%H:%M",
            "%H%M",
            "%H%M%S",
            "%I:%M:%S.%f %p",
            "%I:%M:%S %p",
            "%I:%M %p",
            "%I:%M:%S%p",
            "%I:%M%p",
        ]

        for pattern in patterns:
            try:
                datetime.strptime(value, pattern)
                return True
            except ValueError:
                pass

        return False

    @classmethod
    def is_date_like(
        cls,
        value: Union[str, date, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a date.

        Args:
            value: Value to check (string or date)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is date or string in supported date format

        Examples:
            >>> String.is_date_like('2023-01-01')  # True
            >>> String.is_date_like('01/01/2023')  # True
            >>> String.is_date_like('20230101')    # True
        """
        if not value:
            return False

        if isinstance(value, date):
            return True

        if not isinstance(value, str):
            return False

        tmp_value: str = value.strip() if trim else value

        if re.match(
            r"""^\d{4}[-/.]?\d{2}[-/.]?\d{2}$""", tmp_value
        ) and cls._is_date_like(tmp_value):
            return True

        if re.match(
            r"""^\d{2}[-/.]?\d{2}[-/.]?\d{4}$""", tmp_value
        ) and cls._is_date_like(tmp_value):
            return True

        return False

    @staticmethod
    def _is_datetime_like(
        value: str
    ) -> bool:
        """
        Internal method to check if string matches common datetime formats.

        Args:
            value: String to check

        Returns:
            True if string matches a supported datetime format

        Note:
            Supports multiple datetime formats including:
            - YYYY-MM-DDTHH:MM:SS
            - YYYY/MM/DDTHH:MM:SS
            - YYYY.MM.DDTHH:MM:SS
            - YYYYMMDDHHMMSS
            And their variations with different date component orders and optional microseconds
        """
        patterns: list[str] = [
            "%Y-%m-%dT%H:%M:%S",
            "%Y/%m/%dT%H:%M:%S",
            "%Y.%m.%dT%H:%M:%S",
            "%Y%m%d%H%M%S",
            "%Y-%d-%mT%H:%M:%S",
            "%Y/%d/%mT%H:%M:%S",
            "%Y.%d.%mT%H:%M:%S",
            "%Y%d%m%H%M%S",
            "%m-%d-%YT%H:%M:%S",
            "%m/%d/%YT%H:%M:%S",
            "%m.%d.%YT%H:%M:%S",
            "%m%d%Y%H%M%S",
            "%Y-%m-%dT%H:%M:%S.%f",
            "%Y/%m/%dT%H:%M:%S.%f",
            "%Y.%m.%dT%H:%M:%S.%f",
            "%Y%m%d%H%M%S%f",
            "%Y-%d-%mT%H:%M:%S.%f",
            "%Y/%d/%mT%H:%M:%S.%f",
            "%Y.%d.%mT%H:%M:%S.%f",
            "%Y%d%m%H%M%S%f",
            "%m-%d-%YT%H:%M:%S.%f",
            "%m/%d/%YT%H:%M:%S.%f",
            "%m.%d.%YT%H:%M:%S.%f",
            "%m%d%Y%H%M%S%f",
        ]

        for pattern in patterns:
            try:
                datetime.strptime(value, pattern)
                return True
            except ValueError:
                pass

        return False

    @classmethod
    def is_datetime_like(
        cls,
        value: Union[str, datetime, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a datetime.

        Args:
            value: Value to check (string or datetime)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is datetime or string in supported datetime format

        Examples:
            >>> String.is_datetime_like('2023-01-01T12:00:00')     # True
            >>> String.is_datetime_like('2023-01-01T12:00:00.123') # True
            >>> String.is_datetime_like('2023-01-01')              # False
        """
        if not value:
            return False

        if isinstance(value, datetime):
            return True

        if not isinstance(value, str):
            return False

        tmp_value: str = value.strip() if trim else value

        if re.match(
            r"""^\d{4}[-/.]?\d{2}[-/.]?\d{2}[T]?\d{2}[:]?\d{2}([:]?\d{2}([.]?\d{5})?)?$""",
            tmp_value,
        ) and cls._is_datetime_like(tmp_value):
            return True

        if re.match(
            r"""^\d{2}[-/.]?\d{2}[-/.]?\d{4}[T]?\d{2}[:]?\d{2}([:]?\d{2}([.]?\d{5})?)?$""",
            tmp_value,
        ) and cls._is_datetime_like(tmp_value):
            return True

        return False

    @classmethod
    def is_time_like(
        cls,
        value: Union[str, time, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if value can be interpreted as a time.

        Args:
            value: Value to check (string or time)
            trim: Whether to trim whitespace before checking

        Returns:
            True if value is time or string in supported time format

        Examples:
            >>> String.is_time_like('14:30:00')     # True
            >>> String.is_time_like('14:30:00.123') # True
            >>> String.is_time_like('2:30 PM')      # True
            >>> String.is_time_like('143000')       # True
            >>> String.is_time_like('2023-01-01')   # False
        """
        if not value:
            return False

        if isinstance(value, time):
            return True

        if not isinstance(value, str):
            return False

        tmp_value: str = value.strip() if trim else value

        if re.match(
            r"""^(\d{1,2}):(\d{2})(:(\d{2})([.](\d+))?)?$""", tmp_value
        ) and cls._is_time_like(tmp_value):
            return True

        if re.match(
            r"""^(\d{1,2}):(\d{2})(:(\d{2})([.](\d+))?)?\s*(AM|PM|am|pm)$""", tmp_value
        ) and cls._is_time_like(tmp_value):
            return True

        if re.match(
            r"""^(\d{2})(\d{2})(\d{2})?$""", tmp_value
        ) and cls._is_time_like(tmp_value):
            return True

        return False

    @classmethod
    def to_bool(
        cls,
        value: Union[str, bool, None],
        *,
        default: Union[bool, None] = None,
        trim: bool = True
    ) -> Union[bool, None]:
        """
        Convert value to boolean.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Boolean value or default if conversion fails

        Examples:
            >>> String.to_bool('true')   # True
            >>> String.to_bool('false')  # False
            >>> String.to_bool('yes')    # None
        """
        if isinstance(value, bool):
            return value

        if cls.is_bool_like(value, trim=trim):
            tmp_value: str = value.lower().strip() if trim else value.lower()
            return tmp_value == "true"

        return default

    @classmethod
    def to_float(
        cls,
        value: Union[str, float, None],
        *,
        default: Union[float, None] = None,
        trim: bool = True
    ) -> Union[float, None]:
        """
        Convert value to float.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Float value or default if conversion fails

        Examples:
            >>> String.to_float('1.23')  # 1.23
            >>> String.to_float('-1.23') # -1.23
            >>> String.to_float('abc')   # None
        """
        return float(value) if cls.is_float_like(value, trim=trim) else default

    @classmethod
    def to_int(
        cls,
        value: Union[str, int, None],
        *,
        default: Union[int, None] = None,
        trim: bool = True
    ) -> Union[int, None]:
        """
        Convert value to integer.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Integer value or default if conversion fails

        Examples:
            >>> String.to_int('123')   # 123
            >>> String.to_int('-123')  # -123
            >>> String.to_int('1.23')  # None
        """
        return int(value) if cls.is_int_like(value, trim=trim) else default

    @classmethod
    def to_date(
        cls,
        value: Union[str, date, None],
        *,
        default: Union[date, None] = None,
        trim: bool = True
    ) -> Union[date, None]:
        """
        Convert value to date.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Date value or default if conversion fails

        Examples:
            >>> String.to_date('2023-01-01')  # datetime.date(2023, 1, 1)
            >>> String.to_date('01/01/2023')  # datetime.date(2023, 1, 1)
            >>> String.to_date('invalid')     # None
        """
        if isinstance(value, date):
            return value

        if not cls.is_date_like(value, trim=trim):
            return default

        patterns: list[str] = [
            "%Y-%m-%d",
            "%Y/%m/%d",
            "%Y.%m.%d",
            "%Y%m%d",
            "%Y-%d-%m",
            "%Y/%d/%m",
            "%Y.%d.%m",
            "%Y%d%m",
            "%m-%d-%Y",
            "%m/%d/%Y",
            "%m.%d.%Y",
            "%m%d%Y",
        ]

        dvalue: str = value.strip() if trim else value

        for pattern in patterns:
            try:
                tmp_value = datetime.strptime(dvalue, pattern)
                return tmp_value.date()
            except ValueError:
                pass

        return default

    @classmethod
    def to_datetime(
        cls,
        value: Union[str, datetime, None],
        *,
        default: Union[datetime, None] = None,
        trim: bool = True
    ) -> Union[datetime, None]:
        """
        Convert value to datetime.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Datetime value or default if conversion fails

        Examples:
            >>> String.to_datetime('2023-01-01T12:00:00')     # datetime(2023, 1, 1, 12, 0)
            >>> String.to_datetime('2023-01-01T12:00:00.123') # datetime(2023, 1, 1, 12, 0, 0, 123000)
            >>> String.to_datetime('invalid')                 # None
        """
        if isinstance(value, datetime):
            return value

        if not cls.is_datetime_like(value, trim=trim):
            return default

        patterns: list[str] = [
            "%Y-%m-%dT%H:%M:%S",
            "%Y/%m/%dT%H:%M:%S",
            "%Y.%m.%dT%H:%M:%S",
            "%Y%m%d%H%M%S",
            "%Y-%d-%mT%H:%M:%S",
            "%Y/%d/%mT%H:%M:%S",
            "%Y.%d.%mT%H:%M:%S",
            "%Y%d%m%H%M%S",
            "%m-%d-%YT%H:%M:%S",
            "%m/%d/%YT%H:%M:%S",
            "%m.%d.%YT%H:%M:%S",
            "%m%d%Y%H%M%S",
            "%Y-%m-%dT%H:%M:%S.%f",
            "%Y/%m/%dT%H:%M:%S.%f",
            "%Y.%m.%dT%H:%M:%S.%f",
            "%Y%m%d%H%M%S%f",
            "%Y-%d-%mT%H:%M:%S.%f",
            "%Y/%d/%mT%H:%M:%S.%f",
            "%Y.%d.%mT%H:%M:%S.%f",
            "%Y%d%m%H%M%S%f",
            "%m-%d-%YT%H:%M:%S.%f",
            "%m/%d/%YT%H:%M:%S.%f",
            "%m.%d.%YT%H:%M:%S.%f",
            "%m%d%Y%H%M%S%f",
        ]

        tmp_value: str = value.strip() if trim else value

        for pattern in patterns:
            try:
                tvalue = datetime.strptime(tmp_value, pattern)
                return tvalue
            except ValueError:
                pass

        return default

    @classmethod
    def to_time(
        cls,
        value: Union[str, time, None],
        *,
        default: Union[time, None] = None,
        trim: bool = True
    ) -> Union[time, None]:
        """
        Convert value to time.

        Args:
            value: Value to convert
            default: Default value if conversion fails
            trim: Whether to trim whitespace before converting

        Returns:
            Time value or default if conversion fails

        Examples:
            >>> String.to_time('14:30:00')     # datetime.time(14, 30)
            >>> String.to_time('2:30 PM')      # datetime.time(14, 30)
            >>> String.to_time('143000')       # datetime.time(14, 30, 0)
            >>> String.to_time('invalid')      # None
        """
        if isinstance(value, time):
            return value

        if not cls.is_time_like(value, trim=trim):
            return default

        patterns: list[str] = [
            "%H:%M:%S",
            "%H:%M:%S.%f",
            "%H:%M",
            "%H%M",
            "%H%M%S",
            "%I:%M:%S.%f %p",
            "%I:%M:%S %p",
            "%I:%M %p",
            "%I:%M:%S%p",
            "%I:%M%p",
        ]

        tmp_value: str = value.strip() if trim else value

        for pattern in patterns:
            try:
                tvalue = datetime.strptime(tmp_value, pattern)
                return tvalue.time()
            except ValueError:
                pass

        return default

    @staticmethod
    def has_leading_zero(
        value: Union[str, None],
        *,
        trim: bool = True
    ) -> bool:
        """
        Check if string value has leading zero.

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            True if value starts with '0'

        Examples:
            >>> String.has_leading_zero('01')    # True
            >>> String.has_leading_zero('10')    # False
            >>> String.has_leading_zero(' 01')   # True (with trim=True)
        """
        if value is None:
            return False

        return value.strip().startswith("0") if trim else value.startswith("0")

    @classmethod
    def infer_type(
        cls,
        value: Union[str, bool, int, float, date, time, datetime, None],
        *,
        trim: bool = True
    ) -> DataType:
        """
        Infer the most appropriate data type for a value.

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            DataType enum value representing the inferred type

        Examples:
            >>> String.infer_type('123')           # DataType.INTEGER
            >>> String.infer_type('1.23')          # DataType.FLOAT
            >>> String.infer_type('2023-01-01')    # DataType.DATE
            >>> String.infer_type('true')          # DataType.BOOLEAN
            >>> String.infer_type('abc')           # DataType.STRING
        """
        if cls.is_none_like(value, trim=trim):
            return DataType.NONE

        if cls.is_bool_like(value, trim=trim):
            return DataType.BOOLEAN

        if cls.is_datetime_like(value, trim=trim):
            return DataType.DATETIME

        if cls.is_time_like(value, trim=trim):
            return DataType.TIME

        if cls.is_date_like(value, trim=trim):
            return DataType.DATE

        if cls.is_int_like(value, trim=trim):
            return DataType.INTEGER

        if cls.is_float_like(value, trim=trim):
            return DataType.FLOAT

        if cls.is_empty_like(value, trim=trim):
            return DataType.EMPTY

        return DataType.STRING

    @classmethod
    def infer_type_name(
        cls,
        value: Union[str, bool, int, float, date, time, datetime, None],
        *,
        trim: bool = True
    ) -> str:
        """
        Infer the most appropriate data type name for a value.

        Args:
            value: Value to check
            trim: Whether to trim whitespace before checking

        Returns:
            String name of the inferred type

        Examples:
            >>> String.infer_type_name('123')           # 'INTEGER'
            >>> String.infer_type_name('1.23')          # 'FLOAT'
            >>> String.infer_type_name('2023-01-01')    # 'DATE'
            >>> String.infer_type_name('true')          # 'BOOLEAN'
            >>> String.infer_type_name('abc')           # 'STRING'
        """
        return cls.infer_type(value, trim=trim).name


def profile_values(values: Iterable, *, trim: bool = True) -> DataType:
    """
    Infer the most appropriate data type for a collection of values.

    This function analyzes a collection of values and determines the most
    appropriate data type that can represent all values in the collection.

    Args:
        values: Collection of values to analyze
        trim: Whether to trim whitespace before checking

    Returns:
        DataType enum value representing the inferred type

    Raises:
        ValueError: If values is not iterable

    Examples:
        >>> profile_values(['1', '2', '3'])           # DataType.INTEGER
        >>> profile_values(['1.1', '2.2', '3.3'])     # DataType.FLOAT
        >>> profile_values(['1', '2.2', 'abc'])       # DataType.MIXED
        >>> profile_values(['true', 'false'])         # DataType.BOOLEAN
    """
    types = {
        DataType.BOOLEAN.name: 0,
        DataType.DATE.name: 0,
        DataType.TIME.name: 0,
        DataType.DATETIME.name: 0,
        DataType.INTEGER.name: 0,
        DataType.FLOAT.name: 0,
        DataType.STRING.name: 0,
        DataType.EMPTY.name: 0,
        DataType.NONE.name: 0,
    }

    if not is_iterable_not_string(values):
        raise ValueError("values must be iterable")

    count = 0

    for value in values:
        types[String.infer_type(value, trim=trim).name] += 1
        count += 1

    if types[DataType.EMPTY.name] == count:
        return DataType.EMPTY

    if types[DataType.NONE.name] == count:
        return DataType.NONE

    if types[DataType.NONE.name] + types[DataType.EMPTY.name] == count:
        return DataType.NONE

    if types[DataType.BOOLEAN.name] + types[DataType.EMPTY.name] == count:
        return DataType.BOOLEAN

    if types[DataType.DATE.name] + types[DataType.EMPTY.name] == count:
        return DataType.DATE

    if types[DataType.DATETIME.name] + types[DataType.EMPTY.name] == count:
        return DataType.DATETIME

    if types[DataType.TIME.name] + types[DataType.EMPTY.name] == count:
        return DataType.TIME

    if types[DataType.INTEGER.name] + types[DataType.EMPTY.name] == count:
        return DataType.INTEGER

    if (
        types[DataType.FLOAT.name]
        + types[DataType.INTEGER.name]
        + types[DataType.EMPTY.name]
        == count
    ):
        return DataType.FLOAT

    if types[DataType.STRING.name] + types[DataType.EMPTY.name] == count:
        return DataType.STRING

    return DataType.MIXED


def is_list_like(value: Any) -> bool:
    """
    Check if value behaves like a list.

    Args:
        value: Value to check

    Returns:
        True if value is a list or has list-like behavior (has append, remove, index methods)

    Examples:
        >>> is_list_like([1, 2, 3])        # True
        >>> is_list_like((1, 2, 3))        # False
        >>> is_list_like('abc')            # False
    """
    if isinstance(value, list):
        return True

    if (
        hasattr(value, "__iter__")
        and hasattr(value, "append")
        and hasattr(value, "remove")
        and hasattr(value, "index")
    ):
        return True

    return False


def is_dict_like(value: Any) -> bool:
    """
    Check if value behaves like a dictionary.

    Args:
        value: Value to check

    Returns:
        True if value is a dict or has dict-like behavior (has keys, get, values methods)

    Examples:
        >>> is_dict_like({'a': 1})         # True
        >>> is_dict_like([1, 2, 3])        # False
        >>> is_dict_like('abc')            # False
    """
    if isinstance(value, dict):
        return True

    if hasattr(value, "keys") and hasattr(value, "get") and hasattr(value, "values"):
        return True

    return False


def is_iterable(value: Any) -> bool:
    """
    Check if value is iterable.

    Args:
        value: Value to check

    Returns:
        True if value is iterable (has __iter__, __getitem__, __len__, __next__ methods)

    Examples:
        >>> is_iterable([1, 2, 3])         # True
        >>> is_iterable((1, 2, 3))         # True
        >>> is_iterable('abc')             # True
        >>> is_iterable(123)               # False
    """
    if isinstance(value, (collections.abc.Iterable, typing.Iterable)):
        return True

    if (
        hasattr(value, "__iter__")
        and hasattr(value, "__getitem__")
        and hasattr(value, "__len__")
        and hasattr(value, "__next__")
    ):
        return True

    return False


def is_iterable_not_string(value: Any) -> bool:
    """
    Check if value is iterable but not a string.

    Args:
        value: Value to check

    Returns:
        True if value is iterable and not a string

    Examples:
        >>> is_iterable_not_string([1, 2, 3])  # True
        >>> is_iterable_not_string((1, 2, 3))  # True
        >>> is_iterable_not_string('abc')      # False
        >>> is_iterable_not_string(123)        # False
    """
    if not isinstance(value, str) and is_iterable(value):
        return True

    return False


def is_empty(value: Any) -> bool:
    """
    Check if value is empty.

    Args:
        value: Value to check

    Returns:
        True if value is empty (None, empty string, empty collection)

    Examples:
        >>> is_empty(None)           # True
        >>> is_empty('')             # True
        >>> is_empty('   ')          # True
        >>> is_empty([])             # True
        >>> is_empty({})             # True
        >>> is_empty('abc')          # False
        >>> is_empty([1, 2, 3])      # False
    """
    if value is None:
        return True

    if isinstance(value, str) and not value.strip():
        return True

    if isinstance(value, (list, tuple, set)) and not value:
        return True

    if isinstance(value, dict) and not value:
        return True

    return False



"""
Unit tests for the RandomHelper class.

This module contains comprehensive tests for all methods in the RandomHelper class,
including both secure and non-secure random generation modes.
"""

import unittest
from datetime import date, datetime, timedelta

from splurge_tools.random_helper import RandomHelper


class TestRandomHelper(unittest.TestCase):
    """Test cases for RandomHelper class."""

    def setUp(self):
        """Set up test fixtures."""
        self.helper = RandomHelper()
        self.today = date.today()
        self.now = datetime.now()

    def test_constants(self):
        """Test class constants are correctly defined."""
        self.assertEqual(RandomHelper.INT64_MAX, 2**63 - 1)
        self.assertEqual(RandomHelper.INT64_MIN, -(2**63))
        self.assertEqual(RandomHelper.INT64_MASK, 0x7FFF_FFFF_FFFF_FFFF)
        self.assertEqual(
            len(RandomHelper.ALPHA_CHARS), 52
        )  # 26 lowercase + 26 uppercase
        self.assertEqual(len(RandomHelper.DIGITS), 10)
        self.assertEqual(
            len(RandomHelper.ALPHANUMERIC_CHARS), 62
        )  # 52 letters + 10 digits
        self.assertEqual(len(RandomHelper.BASE58_CHARS), 58)

    def test_as_bytes(self):
        """Test random byte generation."""
        # Test non-secure mode
        bytes1 = RandomHelper.as_bytes(4)
        bytes2 = RandomHelper.as_bytes(4)
        self.assertEqual(len(bytes1), 4)
        self.assertEqual(len(bytes2), 4)
        # Note: bytes1 and bytes2 might be equal by chance, but unlikely

        # Test secure mode
        secure_bytes1 = RandomHelper.as_bytes(4, secure=True)
        secure_bytes2 = RandomHelper.as_bytes(4, secure=True)
        self.assertEqual(len(secure_bytes1), 4)
        self.assertEqual(len(secure_bytes2), 4)

    def test_as_int(self):
        """Test random integer generation."""
        # Test non-secure mode
        int1 = RandomHelper.as_int()
        int2 = RandomHelper.as_int()
        self.assertIsInstance(int1, int)
        self.assertIsInstance(int2, int)
        self.assertTrue(0 <= int1 <= RandomHelper.INT64_MAX)
        self.assertTrue(0 <= int2 <= RandomHelper.INT64_MAX)

        # Test secure mode
        secure_int1 = RandomHelper.as_int(secure=True)
        secure_int2 = RandomHelper.as_int(secure=True)
        self.assertIsInstance(secure_int1, int)
        self.assertIsInstance(secure_int2, int)
        self.assertTrue(0 <= secure_int1 <= RandomHelper.INT64_MAX)
        self.assertTrue(0 <= secure_int2 <= RandomHelper.INT64_MAX)

    def test_as_int_range(self):
        """Test random integer range generation."""
        # Test normal range
        value = RandomHelper.as_int_range(1, 10)
        self.assertTrue(1 <= value <= 10)

        # Test secure mode
        secure_value = RandomHelper.as_int_range(1, 10, secure=True)
        self.assertTrue(1 <= secure_value <= 10)

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_int_range(10, 1)  # lower > upper
        with self.assertRaises(ValueError):
            RandomHelper.as_int_range(RandomHelper.INT64_MIN - 1, 10)  # below min
        with self.assertRaises(ValueError):
            RandomHelper.as_int_range(1, RandomHelper.INT64_MAX + 1)  # above max

    def test_as_float_range(self):
        """Test random float range generation."""
        # Test normal range
        value = RandomHelper.as_float_range(0.0, 1.0)
        self.assertTrue(0.0 <= value <= 1.0)

        # Test negative range
        value = RandomHelper.as_float_range(-1.0, 1.0)
        self.assertTrue(-1.0 <= value <= 1.0)

        # Test edge case
        with self.assertRaises(ValueError):
            RandomHelper.as_float_range(1.0, 0.0)  # lower > upper

    def test_as_string(self):
        """Test random string generation."""
        # Test with custom charset
        value = RandomHelper.as_string(5, "abc")
        self.assertEqual(len(value), 5)
        self.assertTrue(all(c in "abc" for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_string(5, "abc", secure=True)
        self.assertEqual(len(secure_value), 5)
        self.assertTrue(all(c in "abc" for c in secure_value))

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_string(0, "abc")  # length < 1
        with self.assertRaises(ValueError):
            RandomHelper.as_string(5, "")  # empty charset

    def test_as_alpha(self):
        """Test random alphabetic string generation."""
        # Test non-secure mode
        value = RandomHelper.as_alpha(5)
        self.assertEqual(len(value), 5)
        self.assertTrue(all(c in RandomHelper.ALPHA_CHARS for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_alpha(5, secure=True)
        self.assertEqual(len(secure_value), 5)
        self.assertTrue(all(c in RandomHelper.ALPHA_CHARS for c in secure_value))

    def test_as_alphanumeric(self):
        """Test random alphanumeric string generation."""
        # Test non-secure mode
        value = RandomHelper.as_alphanumeric(5)
        self.assertEqual(len(value), 5)
        self.assertTrue(all(c in RandomHelper.ALPHANUMERIC_CHARS for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_alphanumeric(5, secure=True)
        self.assertEqual(len(secure_value), 5)
        self.assertTrue(all(c in RandomHelper.ALPHANUMERIC_CHARS for c in secure_value))

    def test_as_numeric(self):
        """Test random numeric string generation."""
        # Test non-secure mode
        value = RandomHelper.as_numeric(5)
        self.assertEqual(len(value), 5)
        self.assertTrue(all(c in RandomHelper.DIGITS for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_numeric(5, secure=True)
        self.assertEqual(len(secure_value), 5)
        self.assertTrue(all(c in RandomHelper.DIGITS for c in secure_value))

    def test_as_base58(self):
        """Test random Base58 string generation."""
        # Test non-secure mode
        value = RandomHelper.as_base58(5)
        self.assertEqual(len(value), 5)
        self.assertTrue(all(c in RandomHelper.BASE58_CHARS for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_base58(5, secure=True)
        self.assertEqual(len(secure_value), 5)
        self.assertTrue(all(c in RandomHelper.BASE58_CHARS for c in secure_value))

    def test_as_variable_string(self):
        """Test variable length string generation."""
        # Test normal range
        value = RandomHelper.as_variable_string(3, 5, "abc")
        self.assertTrue(3 <= len(value) <= 5)
        self.assertTrue(all(c in "abc" for c in value))

        # Test secure mode
        secure_value = RandomHelper.as_variable_string(3, 5, "abc", secure=True)
        self.assertTrue(3 <= len(secure_value) <= 5)
        self.assertTrue(all(c in "abc" for c in secure_value))

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_variable_string(-1, 5, "abc")  # negative lower bound
        with self.assertRaises(ValueError):
            RandomHelper.as_variable_string(5, 3, "abc")  # lower >= upper

    def test_as_bool(self):
        """Test random boolean generation."""
        # Test non-secure mode
        value = RandomHelper.as_bool()
        self.assertIsInstance(value, bool)

        # Test secure mode
        secure_value = RandomHelper.as_bool(secure=True)
        self.assertIsInstance(secure_value, bool)

    def test_as_masked_string(self):
        """Test masked string generation."""
        # Test with digits and letters
        value = RandomHelper.as_masked_string("###-@@@")
        self.assertEqual(len(value), 7)
        self.assertTrue(all(c in RandomHelper.DIGITS for c in value[:3]))
        self.assertTrue(all(c in RandomHelper.ALPHA_CHARS for c in value[4:]))
        self.assertEqual(value[3], "-")

        # Test secure mode
        secure_value = RandomHelper.as_masked_string("###-@@@", secure=True)
        self.assertEqual(len(secure_value), 7)
        self.assertTrue(all(c in RandomHelper.DIGITS for c in secure_value[:3]))
        self.assertTrue(all(c in RandomHelper.ALPHA_CHARS for c in secure_value[4:]))
        self.assertEqual(secure_value[3], "-")

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_masked_string("")  # empty mask
        with self.assertRaises(ValueError):
            RandomHelper.as_masked_string("---")  # no mask characters

    def test_as_sequenced_string(self):
        """Test sequenced string generation."""
        # Test basic sequence
        values = RandomHelper.as_sequenced_string(3, 3)
        self.assertEqual(len(values), 3)
        self.assertEqual(values, ["000", "001", "002"])

        # Test with prefix and suffix
        values = RandomHelper.as_sequenced_string(3, 3, prefix="ID-", suffix="-END")
        self.assertEqual(len(values), 3)
        self.assertEqual(values, ["ID-000-END", "ID-001-END", "ID-002-END"])

        # Test with start value
        values = RandomHelper.as_sequenced_string(3, 3, start=100)
        self.assertEqual(len(values), 3)
        self.assertEqual(values, ["100", "101", "102"])

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_sequenced_string(0, 3)  # count < 1
        with self.assertRaises(ValueError):
            RandomHelper.as_sequenced_string(3, 0)  # digits < 1
        with self.assertRaises(ValueError):
            RandomHelper.as_sequenced_string(3, 3, start=-1)  # start < 0
        with self.assertRaises(ValueError):
            RandomHelper.as_sequenced_string(1000, 3)  # sequence too long for digits

    def test_as_date(self):
        """Test random date generation."""
        # Test non-secure mode
        value = RandomHelper.as_date(0, 30)
        self.assertIsInstance(value, date)
        self.assertTrue(self.today <= value <= self.today + timedelta(days=30))

        # Test secure mode
        secure_value = RandomHelper.as_date(0, 30, secure=True)
        self.assertIsInstance(secure_value, date)
        self.assertTrue(self.today <= secure_value <= self.today + timedelta(days=30))

        # Test with base_date
        base_date = date(2024, 1, 1)
        value = RandomHelper.as_date(0, 30, base_date=base_date)
        self.assertIsInstance(value, date)
        self.assertTrue(base_date <= value <= base_date + timedelta(days=30))

        # Test negative days
        value = RandomHelper.as_date(-30, 0)
        self.assertIsInstance(value, date)
        self.assertTrue(self.today - timedelta(days=30) <= value <= self.today)

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_date(10, 5)  # lower > upper
        with self.assertRaises(ValueError):
            RandomHelper.as_date(RandomHelper.INT64_MIN - 1, 10)  # below min
        with self.assertRaises(ValueError):
            RandomHelper.as_date(1, RandomHelper.INT64_MAX + 1)  # above max

    def test_as_datetime(self):
        """Test random datetime generation."""
        # Test non-secure mode
        value = RandomHelper.as_datetime(0, 30)
        self.assertIsInstance(value, datetime)
        # Compare only date components
        self.assertTrue(value.date() >= self.now.date())
        self.assertTrue(value.date() <= (self.now + timedelta(days=30)).date())

        # Test secure mode
        secure_value = RandomHelper.as_datetime(0, 30, secure=True)
        self.assertIsInstance(secure_value, datetime)
        # Compare only date components
        self.assertTrue(secure_value.date() >= self.now.date())
        self.assertTrue(secure_value.date() <= (self.now + timedelta(days=30)).date())

        # Test with base_date
        base_date = datetime(2024, 1, 1, 12, 0, 0)
        value = RandomHelper.as_datetime(0, 30, base_date=base_date)
        self.assertIsInstance(value, datetime)
        # Compare only date components
        self.assertTrue(value.date() >= base_date.date())
        self.assertTrue(value.date() <= (base_date + timedelta(days=30)).date())
        # Verify time components are randomized
        self.assertIsInstance(value.hour, int)
        self.assertIsInstance(value.minute, int)
        self.assertIsInstance(value.second, int)
        self.assertIsInstance(value.microsecond, int)
        self.assertTrue(0 <= value.hour <= 23)
        self.assertTrue(0 <= value.minute <= 59)
        self.assertTrue(0 <= value.second <= 59)
        self.assertTrue(0 <= value.microsecond <= 999999)

        # Test negative days
        value = RandomHelper.as_datetime(-30, 0)
        self.assertIsInstance(value, datetime)
        # Compare only date components
        self.assertTrue(value.date() >= (self.now - timedelta(days=30)).date())
        self.assertTrue(value.date() <= self.now.date())

        # Test edge cases
        with self.assertRaises(ValueError):
            RandomHelper.as_datetime(10, 5)  # lower > upper
        with self.assertRaises(ValueError):
            RandomHelper.as_datetime(RandomHelper.INT64_MIN - 1, 10)  # below min
        with self.assertRaises(ValueError):
            RandomHelper.as_datetime(1, RandomHelper.INT64_MAX + 1)  # above max

        # Test datetime components
        value = RandomHelper.as_datetime(0, 1)
        self.assertIsInstance(value.hour, int)
        self.assertIsInstance(value.minute, int)
        self.assertIsInstance(value.second, int)
        self.assertIsInstance(value.microsecond, int)
        self.assertTrue(0 <= value.hour <= 23)
        self.assertTrue(0 <= value.minute <= 59)
        self.assertTrue(0 <= value.second <= 59)
        self.assertTrue(0 <= value.microsecond <= 999999)


if __name__ == "__main__":
    unittest.main()



"""Unit tests for DataTransformer class."""

import unittest
from statistics import mean

from splurge_tools.data_transformer import DataTransformer
from splurge_tools.tabular_data_model import TabularDataModel


class TestDataTransformer(unittest.TestCase):
    """Test cases for DataTransformer class."""

    def setUp(self):
        """Set up test data."""
        # Sample data for testing
        self.sample_data = [
            ["Name", "Category", "Value", "Date"],
            ["John", "A", "10", "2024-01-01"],
            ["John", "B", "20", "2024-01-01"],
            ["Jane", "A", "15", "2024-01-01"],
            ["Jane", "B", "25", "2024-01-01"],
            ["Bob", "A", "12", "2024-01-02"],
            ["Bob", "B", "22", "2024-01-02"],
        ]
        self.model = TabularDataModel(self.sample_data)
        self.transformer = DataTransformer(self.model)

        # Data with duplicates for testing
        self.duplicate_data = [
            ["Name", "Category", "Value", "Date"],
            ["John", "A", "10", "2024-01-01"],
            ["John", "A", "15", "2024-01-01"],  # Duplicate Category for John
            ["John", "B", "20", "2024-01-01"],
            ["Jane", "A", "15", "2024-01-01"],
            ["Jane", "B", "25", "2024-01-01"],
            ["Jane", "B", "30", "2024-01-01"],  # Duplicate Category for Jane
            ["Bob", "A", "12", "2024-01-02"],
            ["Bob", "B", "22", "2024-01-02"],
        ]
        self.duplicate_model = TabularDataModel(self.duplicate_data)
        self.duplicate_transformer = DataTransformer(self.duplicate_model)

    def test_pivot(self):
        """Test pivot operation."""
        # Pivot by Name and Category
        result = self.transformer.pivot(
            index_cols=["Name"], columns_col="Category", values_col="Value"
        )

        # Check header
        self.assertEqual(result.column_names, ["Name", "A", "B"])

        # Check data
        rows = list(result.iter_rows())
        self.assertEqual(len(rows), 3)  # 3 names

        # Check John's values
        john_row = next(row for row in rows if row["Name"] == "John")
        self.assertEqual(john_row["A"], "10")
        self.assertEqual(john_row["B"], "20")

    def test_melt(self):
        """Test melt operation."""
        # Melt Value and Date columns
        result = self.transformer.melt(
            id_vars=["Name", "Category"], value_vars=["Value", "Date"]
        )

        # Check header
        self.assertEqual(result.column_names, ["Name", "Category", "variable", "value"])

        # Check data
        rows = list(result.iter_rows())
        self.assertEqual(len(rows), 12)  # 6 data rows * 2 variables

        # Check first row
        self.assertEqual(rows[0]["Name"], "John")
        self.assertEqual(rows[0]["Category"], "A")
        self.assertEqual(rows[0]["variable"], "Value")
        self.assertEqual(rows[0]["value"], "10")

    def test_group_by(self):
        """Test group by operation."""
        # Group by Name and calculate mean Value
        result = self.transformer.group_by(
            group_cols=["Name"], agg_dict={"Value": lambda x: mean(float(v) for v in x)}
        )

        # Check header
        self.assertEqual(result.column_names, ["Name", "Value"])

        # Check data
        rows = list(result.iter_rows())
        self.assertEqual(len(rows), 3)  # 3 names

        # Check John's mean value
        john_row = next(row for row in rows if row["Name"] == "John")
        self.assertEqual(float(john_row["Value"]), 15.0)  # (10 + 20) / 2

    def test_transform_column(self):
        """Test column transformation."""
        # Double the Value column
        result = self.transformer.transform_column(
            column="Value", transform_func=lambda x: str(float(x) * 2)
        )

        # Check data
        rows = list(result.iter_rows())
        self.assertEqual(len(rows), 6)  # 6 data rows

        # Check first row
        self.assertEqual(rows[0]["Value"], "20.0")  # 10 * 2

    def test_pivot_with_duplicates(self):
        """Test pivot operation with duplicate values."""
        # Test that pivot raises error when duplicates are found without agg_func
        with self.assertRaises(ValueError) as context:
            self.duplicate_transformer.pivot(
                index_cols=["Name"], columns_col="Category", values_col="Value"
            )
        self.assertIn("Duplicate values found", str(context.exception))

        # Test pivot with mean aggregation
        result = self.duplicate_transformer.pivot(
            index_cols=["Name"],
            columns_col="Category",
            values_col="Value",
            agg_func=lambda x: mean(float(v) for v in x),
        )

        # Check header
        self.assertEqual(result.column_names, ["Name", "A", "B"])

        # Check data
        rows = list(result.iter_rows())
        self.assertEqual(len(rows), 3)  # 3 names

        # Check John's values (A should be mean of 10 and 15)
        john_row = next(row for row in rows if row["Name"] == "John")
        self.assertEqual(float(john_row["A"]), 12.5)  # (10 + 15) / 2
        self.assertEqual(float(john_row["B"]), 20.0)  # No duplicates

        # Check Jane's values (B should be mean of 25 and 30)
        jane_row = next(row for row in rows if row["Name"] == "Jane")
        self.assertEqual(float(jane_row["A"]), 15.0)  # No duplicates
        self.assertEqual(float(jane_row["B"]), 27.5)  # (25 + 30) / 2

        # Check Bob's values (no duplicates)
        bob_row = next(row for row in rows if row["Name"] == "Bob")
        self.assertEqual(float(bob_row["A"]), 12.0)
        self.assertEqual(float(bob_row["B"]), 22.0)

    def test_pivot_with_duplicates_and_custom_agg(self):
        """Test pivot operation with duplicate values and custom aggregation."""
        # Test pivot with max aggregation
        result = self.duplicate_transformer.pivot(
            index_cols=["Name"],
            columns_col="Category",
            values_col="Value",
            agg_func=lambda x: max(float(v) for v in x),
        )

        # Check data
        rows = list(result.iter_rows())

        # Check John's values (A should be max of 10 and 15)
        john_row = next(row for row in rows if row["Name"] == "John")
        self.assertEqual(float(john_row["A"]), 15.0)  # max(10, 15)
        self.assertEqual(float(john_row["B"]), 20.0)  # No duplicates

        # Check Jane's values (B should be max of 25 and 30)
        jane_row = next(row for row in rows if row["Name"] == "Jane")
        self.assertEqual(float(jane_row["A"]), 15.0)  # No duplicates
        self.assertEqual(float(jane_row["B"]), 30.0)  # max(25, 30)

        # Test pivot with min aggregation
        result = self.duplicate_transformer.pivot(
            index_cols=["Name"],
            columns_col="Category",
            values_col="Value",
            agg_func=lambda x: min(float(v) for v in x),
        )

        # Check data
        rows = list(result.iter_rows())

        # Check John's values (A should be min of 10 and 15)
        john_row = next(row for row in rows if row["Name"] == "John")
        self.assertEqual(float(john_row["A"]), 10.0)  # min(10, 15)
        self.assertEqual(float(john_row["B"]), 20.0)  # No duplicates

        # Check Jane's values (B should be min of 25 and 30)
        jane_row = next(row for row in rows if row["Name"] == "Jane")
        self.assertEqual(float(jane_row["A"]), 15.0)  # No duplicates
        self.assertEqual(float(jane_row["B"]), 25.0)  # min(25, 30)



import os
import tempfile
import unittest

from splurge_tools.text_file_helper import TextFileHelper


class TestTextFileHelper(unittest.TestCase):
    def setUp(self):
        # Create a temporary file for testing
        self.temp_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
        self.test_content = [
            "Line 1",
            "Line 2",
            "Line 3",
            "  Line 4 with spaces  ",
            "Line 5",
        ]
        self.temp_file.write("\n".join(self.test_content))
        self.temp_file.close()

    def tearDown(self):
        # Clean up the temporary file
        os.unlink(self.temp_file.name)

    def test_line_count(self):
        """Test line counting functionality"""
        # Test normal case
        self.assertEqual(TextFileHelper.line_count(self.temp_file.name), 5)

        # Test empty file
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as empty_file:
            empty_file.write("")
        self.assertEqual(TextFileHelper.line_count(empty_file.name), 0)
        os.unlink(empty_file.name)

        # Test file not found
        with self.assertRaises(FileNotFoundError):
            TextFileHelper.line_count("nonexistent_file.txt")

        # Test with different encoding
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-16", delete=False
        ) as encoded_file:
            encoded_file.write("Line 1\nLine 2")
        self.assertEqual(
            TextFileHelper.line_count(encoded_file.name, encoding="utf-16"), 2
        )
        os.unlink(encoded_file.name)

    def test_preview(self):
        """Test file preview functionality"""
        # Test normal case with default parameters
        preview_lines = TextFileHelper.preview(self.temp_file.name)
        self.assertEqual(len(preview_lines), 5)
        self.assertEqual(preview_lines[0], "Line 1")
        self.assertEqual(preview_lines[3], "Line 4 with spaces")

        # Test with strip=False
        preview_lines = TextFileHelper.preview(self.temp_file.name, strip=False)
        self.assertEqual(preview_lines[3], "  Line 4 with spaces  ")

        # Test with max_lines limit
        preview_lines = TextFileHelper.preview(self.temp_file.name, max_lines=3)
        self.assertEqual(len(preview_lines), 3)
        self.assertEqual(preview_lines[2], "Line 3")

        # Test with different encoding
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-16", delete=False
        ) as encoded_file:
            encoded_file.write("Line 1\nLine 2")
        preview_lines = TextFileHelper.preview(encoded_file.name, encoding="utf-16")
        self.assertEqual(preview_lines, ["Line 1", "Line 2"])
        os.unlink(encoded_file.name)

        # Test invalid max_lines
        with self.assertRaises(ValueError):
            TextFileHelper.preview(self.temp_file.name, max_lines=0)

        # Test file not found
        with self.assertRaises(FileNotFoundError):
            TextFileHelper.preview("nonexistent_file.txt")

    def test_load(self):
        """Test file loading functionality"""
        # Test normal case with default parameters (strip=True)
        loaded_lines = TextFileHelper.load(self.temp_file.name)
        self.assertEqual(len(loaded_lines), 5)
        self.assertEqual(loaded_lines[0], "Line 1")
        self.assertEqual(loaded_lines[3], "Line 4 with spaces")

        # Test with strip=False
        loaded_lines = TextFileHelper.load(self.temp_file.name, strip=False)
        self.assertEqual(loaded_lines[3], "  Line 4 with spaces  ")

        # Test with skip_header_rows
        loaded_lines = TextFileHelper.load(self.temp_file.name, skip_header_rows=2)
        self.assertEqual(len(loaded_lines), 3)
        self.assertEqual(loaded_lines[0], "Line 3")

        # Test with skip_footer_rows
        loaded_lines = TextFileHelper.load(self.temp_file.name, skip_footer_rows=2)
        self.assertEqual(len(loaded_lines), 3)
        self.assertEqual(loaded_lines[-1], "Line 3")

        # Test with both skip_header_rows and skip_footer_rows
        loaded_lines = TextFileHelper.load(
            self.temp_file.name, skip_header_rows=1, skip_footer_rows=1
        )
        self.assertEqual(len(loaded_lines), 3)
        self.assertEqual(loaded_lines[0], "Line 2")
        self.assertEqual(loaded_lines[-1], "Line 4 with spaces")

        # Test with different encoding
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-16", delete=False
        ) as encoded_file:
            encoded_file.write("Line 1\nLine 2")
        loaded_lines = TextFileHelper.load(encoded_file.name, encoding="utf-16")
        self.assertEqual(loaded_lines, ["Line 1", "Line 2"])
        os.unlink(encoded_file.name)

        # Test empty file
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as empty_file:
            empty_file.write("")
        self.assertEqual(TextFileHelper.load(empty_file.name), [])
        os.unlink(empty_file.name)

        # Test file not found
        with self.assertRaises(FileNotFoundError):
            TextFileHelper.load("nonexistent_file.txt")


if __name__ == "__main__":
    unittest.main()



"""
Unit tests for type_helper module
"""

import unittest
from datetime import date, datetime, time

from splurge_tools.type_helper import (
    DataType,
    String,
    is_dict_like,
    is_empty,
    is_iterable,
    is_iterable_not_string,
    is_list_like,
    profile_values,
)


class TestString(unittest.TestCase):
    """Test cases for String class methods"""

    def test_is_bool_like(self):
        """Test boolean-like value detection"""
        # Test boolean values
        self.assertTrue(String.is_bool_like(True))
        self.assertTrue(String.is_bool_like(False))

        # Test string values
        self.assertTrue(String.is_bool_like("true"))
        self.assertTrue(String.is_bool_like("false"))
        self.assertTrue(String.is_bool_like("TRUE"))
        self.assertTrue(String.is_bool_like("FALSE"))

        # Test with whitespace
        self.assertTrue(String.is_bool_like(" true "))
        self.assertTrue(String.is_bool_like(" false "))

        # Test non-boolean values
        self.assertFalse(String.is_bool_like("yes"))
        self.assertFalse(String.is_bool_like("no"))
        self.assertFalse(String.is_bool_like(None))
        self.assertFalse(String.is_bool_like(1))
        self.assertFalse(String.is_bool_like(0))

    def test_is_none_like(self):
        """Test None-like value detection"""
        # Test None values
        self.assertTrue(String.is_none_like(None))

        # Test string values
        self.assertTrue(String.is_none_like("none"))
        self.assertTrue(String.is_none_like("null"))
        self.assertTrue(String.is_none_like("NONE"))
        self.assertTrue(String.is_none_like("NULL"))

        # Test with whitespace
        self.assertTrue(String.is_none_like(" none "))
        self.assertTrue(String.is_none_like(" null "))

        # Test non-None values
        self.assertFalse(String.is_none_like(""))
        self.assertFalse(String.is_none_like("0"))
        self.assertFalse(String.is_none_like(0))
        self.assertFalse(String.is_none_like([]))

    def test_is_float_like(self):
        """Test float-like value detection"""
        # Test float values
        self.assertTrue(String.is_float_like(1.23))
        self.assertTrue(String.is_float_like(-1.23))

        # Test string values
        self.assertTrue(String.is_float_like("1.23"))
        self.assertTrue(String.is_float_like("-1.23"))
        self.assertTrue(String.is_float_like(".23"))
        self.assertTrue(String.is_float_like("1."))

        # Test with whitespace
        self.assertTrue(String.is_float_like(" 1.23 "))

        # Test non-float values
        self.assertFalse(String.is_float_like("1,23"))
        self.assertFalse(String.is_float_like("abc"))
        self.assertFalse(String.is_float_like(None))
        self.assertFalse(String.is_float_like([]))

    def test_is_int_like(self):
        """Test integer-like value detection"""
        # Test integer values
        self.assertTrue(String.is_int_like(123))
        self.assertTrue(String.is_int_like(-123))

        # Test string values
        self.assertTrue(String.is_int_like("123"))
        self.assertTrue(String.is_int_like("-123"))

        # Test with whitespace
        self.assertTrue(String.is_int_like(" 123 "))

        # Test non-integer values
        self.assertFalse(String.is_int_like("123.45"))
        self.assertFalse(String.is_int_like("abc"))
        self.assertFalse(String.is_int_like(None))
        self.assertFalse(String.is_int_like([]))

    def test_is_numeric_like(self):
        """Test numeric-like value detection"""
        # Test numeric values
        self.assertTrue(String.is_numeric_like(123))
        self.assertTrue(String.is_numeric_like(123.45))
        self.assertTrue(String.is_numeric_like(-123))
        self.assertTrue(String.is_numeric_like(-123.45))

        # Test string values
        self.assertTrue(String.is_numeric_like("123"))
        self.assertTrue(String.is_numeric_like("123.45"))
        self.assertTrue(String.is_numeric_like("-123"))
        self.assertTrue(String.is_numeric_like("-123.45"))

        # Test non-numeric values
        self.assertFalse(String.is_numeric_like("abc"))
        self.assertFalse(String.is_numeric_like(None))
        self.assertFalse(String.is_numeric_like([]))

    def test_is_date_like(self):
        """Test date-like value detection"""
        # Test date values
        test_date = date(2023, 1, 1)
        self.assertTrue(String.is_date_like(test_date))

        # Test string values
        self.assertTrue(String.is_date_like("2023-01-01"))
        self.assertTrue(String.is_date_like("2023/01/01"))
        self.assertTrue(String.is_date_like("01-01-2023"))
        self.assertTrue(String.is_date_like("01/01/2023"))

        # Test with whitespace
        self.assertTrue(String.is_date_like(" 2023-01-01 "))

        # Test non-date values
        self.assertFalse(String.is_date_like("2023-13-27"))  # Invalid month
        self.assertFalse(String.is_date_like("abc"))
        self.assertFalse(String.is_date_like(None))
        self.assertFalse(String.is_date_like([]))

    def test_is_datetime_like(self):
        """Test datetime-like value detection"""
        # Test datetime values
        test_datetime = datetime(2023, 1, 1, 12, 30, 45)
        self.assertTrue(String.is_datetime_like(test_datetime))

        # Test string values
        self.assertTrue(String.is_datetime_like("2023-01-01T12:30:45"))
        self.assertTrue(String.is_datetime_like("2023-01-01T12:30:45.12340"))
        self.assertTrue(String.is_datetime_like("2023/01/01T12:30:45"))

        # Test with whitespace
        self.assertTrue(String.is_datetime_like(" 2023-01-01T12:30:45 "))

        # Test non-datetime values
        self.assertFalse(
            String.is_datetime_like("2023-13-27T12:30:45")
        )  # Invalid month
        self.assertFalse(String.is_datetime_like("abc"))
        self.assertFalse(String.is_datetime_like(None))
        self.assertFalse(String.is_datetime_like([]))

    def test_to_bool(self):
        """Test boolean conversion"""
        # Test boolean values
        self.assertTrue(String.to_bool(True))
        self.assertFalse(String.to_bool(False))

        # Test string values
        self.assertTrue(String.to_bool("true"))
        self.assertFalse(String.to_bool("false"))

        # Test with default
        self.assertIsNone(String.to_bool("invalid", default=None))
        self.assertFalse(String.to_bool("invalid", default=False))

        # Test with whitespace
        self.assertTrue(String.to_bool(" true "))
        self.assertFalse(String.to_bool(" false "))

    def test_to_float(self):
        """Test float conversion"""
        # Test float values
        self.assertEqual(String.to_float(1.23), 1.23)
        self.assertEqual(String.to_float(-1.23), -1.23)

        # Test string values
        self.assertEqual(String.to_float("1.23"), 1.23)
        self.assertEqual(String.to_float("-1.23"), -1.23)

        # Test with default
        self.assertIsNone(String.to_float("invalid", default=None))
        self.assertEqual(String.to_float("invalid", default=0.0), 0.0)

        # Test with whitespace
        self.assertEqual(String.to_float(" 1.23 "), 1.23)

    def test_to_int(self):
        """Test integer conversion"""
        # Test integer values
        self.assertEqual(String.to_int(123), 123)
        self.assertEqual(String.to_int(-123), -123)

        # Test string values
        self.assertEqual(String.to_int("123"), 123)
        self.assertEqual(String.to_int("-123"), -123)

        # Test with default
        self.assertIsNone(String.to_int("invalid", default=None))
        self.assertEqual(String.to_int("invalid", default=0), 0)

        # Test with whitespace
        self.assertEqual(String.to_int(" 123 "), 123)

    def test_to_date(self):
        """Test date conversion"""
        # Test date values
        test_date = date(2023, 1, 1)
        self.assertEqual(String.to_date(test_date), test_date)

        # Test string values
        self.assertEqual(String.to_date("2023-01-01"), date(2023, 1, 1))
        self.assertEqual(String.to_date("2023/01/01"), date(2023, 1, 1))

        # Test with default
        self.assertIsNone(String.to_date("invalid", default=None))
        default_date = date(2023, 1, 1)
        self.assertEqual(String.to_date("invalid", default=default_date), default_date)

        # Test with whitespace
        self.assertEqual(String.to_date(" 2023-01-01 "), date(2023, 1, 1))

    def test_to_datetime(self):
        """Test datetime conversion"""
        # Test datetime values
        test_datetime = datetime(2023, 1, 1, 12, 30, 45)
        self.assertEqual(String.to_datetime(test_datetime), test_datetime)

        # Test string values
        self.assertEqual(
            String.to_datetime("2023-01-01T12:30:45"), datetime(2023, 1, 1, 12, 30, 45)
        )

        # Test with default
        self.assertIsNone(String.to_datetime("invalid", default=None))
        default_datetime = datetime(2023, 1, 1, 12, 30, 45)
        self.assertEqual(
            String.to_datetime("invalid", default=default_datetime), default_datetime
        )

        # Test with whitespace
        self.assertEqual(
            String.to_datetime(" 2023-01-01T12:30:45 "),
            datetime(2023, 1, 1, 12, 30, 45),
        )

    def test_has_leading_zero(self):
        """Test leading zero detection"""
        # Test with leading zero
        self.assertTrue(String.has_leading_zero("01"))
        self.assertTrue(String.has_leading_zero(" 01 "))

        # Test without leading zero
        self.assertFalse(String.has_leading_zero("1"))
        self.assertFalse(String.has_leading_zero("10"))
        self.assertFalse(String.has_leading_zero(None))
        self.assertFalse(String.has_leading_zero(""))

    def test_infer_type(self):
        """Test type inference"""
        # Test basic types
        self.assertEqual(String.infer_type(None), DataType.NONE)
        self.assertEqual(String.infer_type(True), DataType.BOOLEAN)
        self.assertEqual(String.infer_type(123), DataType.INTEGER)
        self.assertEqual(String.infer_type(123.45), DataType.FLOAT)
        self.assertEqual(String.infer_type("abc"), DataType.STRING)

        # Test date types
        self.assertEqual(String.infer_type(date(2023, 1, 1)), DataType.DATE)
        self.assertEqual(
            String.infer_type(datetime(2023, 1, 1, 12, 30, 45)), DataType.DATETIME
        )
        self.assertEqual(String.infer_type(time(14, 30, 45)), DataType.TIME)

        # Test string representations
        self.assertEqual(String.infer_type("2023-01-01"), DataType.DATE)
        self.assertEqual(String.infer_type("2023-01-01T12:30:45"), DataType.DATETIME)
        self.assertEqual(String.infer_type("14:30:45"), DataType.TIME)
        self.assertEqual(String.infer_type("2:30 PM"), DataType.TIME)
        self.assertEqual(String.infer_type("123"), DataType.INTEGER)
        self.assertEqual(String.infer_type("123.45"), DataType.FLOAT)
        self.assertEqual(String.infer_type("true"), DataType.BOOLEAN)

    def test_is_empty_like(self):
        """Test is_empty_like method."""
        # Test empty strings
        self.assertTrue(String.is_empty_like(""))
        self.assertTrue(String.is_empty_like("   "))
        self.assertTrue(String.is_empty_like("\t\n\r"))

        # Test non-empty strings
        self.assertFalse(String.is_empty_like("abc"))
        self.assertFalse(String.is_empty_like("  abc  "))

        # Test non-string values
        self.assertFalse(String.is_empty_like(None))
        self.assertFalse(String.is_empty_like(123))
        self.assertFalse(String.is_empty_like([]))
        self.assertFalse(String.is_empty_like({}))

        # Test with trim=False
        self.assertFalse(String.is_empty_like("   ", trim=False))
        self.assertTrue(String.is_empty_like("", trim=False))

    def test_is_time_like(self):
        """Test time-like value detection"""
        # Test time values
        test_time = time(14, 30, 45)
        self.assertTrue(String.is_time_like(test_time))

        # Test valid time strings - 24-hour format
        self.assertTrue(String.is_time_like("14:30:45"))
        self.assertTrue(String.is_time_like("14:30:45.123456"))
        self.assertTrue(String.is_time_like("14:30"))
        self.assertTrue(String.is_time_like("143045"))
        self.assertTrue(String.is_time_like("1430"))
        self.assertTrue(String.is_time_like("00:00:00"))  # Midnight
        self.assertTrue(String.is_time_like("23:59:59"))  # End of day
        self.assertTrue(String.is_time_like("12:00:00"))  # Noon

        # Test valid time strings - 12-hour format
        self.assertTrue(String.is_time_like("2:30 PM"))
        self.assertTrue(String.is_time_like("2:30:45 PM"))
        self.assertTrue(String.is_time_like("2:30PM"))
        self.assertTrue(String.is_time_like("2:30:45PM"))
        self.assertTrue(String.is_time_like("12:00 AM"))  # Midnight
        self.assertTrue(String.is_time_like("12:00 PM"))  # Noon
        self.assertTrue(String.is_time_like("11:59 PM"))  # End of day
        self.assertTrue(String.is_time_like("12:30 AM"))  # Early morning

        # Test with whitespace
        self.assertTrue(String.is_time_like(" 14:30:45 "))
        self.assertTrue(String.is_time_like(" 2:30 PM "))

        # Test invalid time values
        self.assertFalse(String.is_time_like("25:30:45"))  # Invalid hour
        self.assertFalse(String.is_time_like("14:60:45"))  # Invalid minute
        self.assertFalse(String.is_time_like("14:30:60"))  # Invalid second
        self.assertFalse(String.is_time_like("13:30 PM"))  # Invalid 13 PM
        self.assertFalse(String.is_time_like("0:30 AM"))   # Invalid 0 AM
        self.assertFalse(String.is_time_like("12:30:60 PM"))  # Invalid seconds
        self.assertFalse(String.is_time_like("abc"))
        self.assertFalse(String.is_time_like(None))
        self.assertFalse(String.is_time_like([]))
        self.assertFalse(String.is_time_like("2023-01-01"))  # Date, not time
        self.assertFalse(String.is_time_like("14:30:45:67"))  # Too many components
        self.assertFalse(String.is_time_like("14:30:45.123456789"))  # Too many microseconds

    def test_to_time(self):
        """Test time conversion"""
        # Test time values
        test_time = time(14, 30, 45)
        self.assertEqual(String.to_time(test_time), test_time)

        # Test valid time strings - 24-hour format
        self.assertEqual(
            String.to_time("14:30:45"), time(14, 30, 45)
        )
        self.assertEqual(
            String.to_time("14:30"), time(14, 30)
        )
        self.assertEqual(
            String.to_time("143045"), time(14, 30, 45)
        )
        self.assertEqual(
            String.to_time("1430"), time(14, 30)
        )
        self.assertEqual(
            String.to_time("00:00:00"), time(0, 0, 0)  # Midnight
        )
        self.assertEqual(
            String.to_time("23:59:59"), time(23, 59, 59)  # End of day
        )
        self.assertEqual(
            String.to_time("12:00:00"), time(12, 0, 0)  # Noon
        )

        # Test valid time strings - 12-hour format
        self.assertEqual(
            String.to_time("2:30 PM"), time(14, 30)
        )
        self.assertEqual(
            String.to_time("2:30:45 PM"), time(14, 30, 45)
        )
        self.assertEqual(
            String.to_time("12:00 AM"), time(0, 0, 0)  # Midnight
        )
        self.assertEqual(
            String.to_time("12:00 PM"), time(12, 0, 0)  # Noon
        )
        self.assertEqual(
            String.to_time("11:59 PM"), time(23, 59)  # End of day
        )
        self.assertEqual(
            String.to_time("12:30 AM"), time(0, 30)  # Early morning
        )

        # Test with microseconds
        self.assertEqual(
            String.to_time("14:30:45.123456"), time(14, 30, 45, 123456)
        )
        self.assertEqual(
            String.to_time("2:30:45.123456 PM"), time(14, 30, 45, 123456)
        )

        # Test with default
        self.assertIsNone(String.to_time("invalid", default=None))
        default_time = time(12, 0, 0)
        self.assertEqual(
            String.to_time("invalid", default=default_time), default_time
        )

        # Test with whitespace
        self.assertEqual(
            String.to_time(" 14:30:45 "),
            time(14, 30, 45),
        )
        self.assertEqual(
            String.to_time(" 2:30 PM "),
            time(14, 30),
        )

        # Test edge cases
        self.assertEqual(
            String.to_time("00:00"), time(0, 0)
        )
        self.assertEqual(
            String.to_time("23:59"), time(23, 59)
        )

    def test_time_type_inference(self):
        """Test time type inference and edge cases"""
        # Test time type inference
        self.assertEqual(String.infer_type(time(14, 30, 45)), DataType.TIME)
        self.assertEqual(String.infer_type("14:30:45"), DataType.TIME)
        self.assertEqual(String.infer_type("2:30 PM"), DataType.TIME)
        self.assertEqual(String.infer_type("143045"), DataType.TIME)
        
        # Test time type name inference
        self.assertEqual(String.infer_type_name(time(14, 30, 45)), "TIME")
        self.assertEqual(String.infer_type_name("14:30:45"), "TIME")
        self.assertEqual(String.infer_type_name("2:30 PM"), "TIME")
        
        # Test boundary conditions
        self.assertTrue(String.is_time_like("00:00:00"))  # Start of day
        self.assertTrue(String.is_time_like("23:59:59"))  # End of day
        self.assertTrue(String.is_time_like("12:00:00"))  # Noon
        self.assertTrue(String.is_time_like("12:00:00.000000"))  # Noon with microseconds
        
        # Test 12-hour format boundaries
        self.assertTrue(String.is_time_like("12:00 AM"))  # Midnight
        self.assertTrue(String.is_time_like("12:00 PM"))  # Noon
        self.assertTrue(String.is_time_like("11:59 PM"))  # End of day
        self.assertTrue(String.is_time_like("12:01 AM"))  # After midnight
        
        # Test invalid boundary conditions
        self.assertFalse(String.is_time_like("24:00:00"))  # Invalid hour
        self.assertFalse(String.is_time_like("23:60:00"))  # Invalid minute
        self.assertFalse(String.is_time_like("23:59:60"))  # Invalid second
        self.assertFalse(String.is_time_like("13:00 PM"))  # Invalid 13 PM
        self.assertFalse(String.is_time_like("0:00 AM"))   # Invalid 0 AM
        
        # Test conversion edge cases
        self.assertEqual(String.to_time("00:00:00.000000"), time(0, 0, 0, 0))
        self.assertEqual(String.to_time("23:59:59.999999"), time(23, 59, 59, 999999))
        
        # Test with trim=False
        self.assertFalse(String.is_time_like(" 14:30:45 ", trim=False))
        self.assertFalse(String.is_time_like(" 2:30 PM ", trim=False))


class TestProfileValues(unittest.TestCase):
    """Test cases for profile_values function"""

    def test_profile_values(self):
        """Test profile_values function."""
        # Test empty collections
        self.assertEqual(profile_values([]), DataType.EMPTY)

        # Test empty strings
        self.assertEqual(profile_values(["", "   ", "\t"]), DataType.EMPTY)

        # Test None values
        self.assertEqual(profile_values([None, None]), DataType.NONE)

        # Test mixed None and empty
        self.assertEqual(profile_values([None, "", "   "]), DataType.NONE)

        # Test boolean values
        self.assertEqual(profile_values(["true", "false"]), DataType.BOOLEAN)
        self.assertEqual(profile_values(["true", "false", ""]), DataType.BOOLEAN)

        # Test date values
        self.assertEqual(profile_values(["2023-01-01", "2023-01-02"]), DataType.DATE)
        self.assertEqual(
            profile_values(["2023-01-01", "2023-01-02", ""]), DataType.DATE
        )

        # Test datetime values
        self.assertEqual(
            profile_values(["2023-01-01T12:00:00", "2023-01-02T12:00:00"]),
            DataType.DATETIME,
        )
        self.assertEqual(
            profile_values(["2023-01-01T12:00:00", "2023-01-02T12:00:00", ""]),
            DataType.DATETIME,
        )

        # Test time values
        self.assertEqual(
            profile_values(["14:30:00", "15:45:00"]),
            DataType.TIME,
        )
        self.assertEqual(
            profile_values(["14:30:00", "15:45:00", ""]),
            DataType.TIME,
        )
        self.assertEqual(
            profile_values(["2:30 PM", "3:45 PM"]),
            DataType.TIME,
        )
        self.assertEqual(
            profile_values(["143000", "154500"]),
            DataType.TIME,
        )
        self.assertEqual(
            profile_values(["00:00:00", "23:59:59"]),
            DataType.TIME,
        )
        self.assertEqual(
            profile_values(["12:00 AM", "12:00 PM"]),
            DataType.TIME,
        )

        # Test integer values
        self.assertEqual(profile_values(["1", "2", "3"]), DataType.INTEGER)
        self.assertEqual(profile_values(["1", "2", "3", ""]), DataType.INTEGER)

        # Test float values
        self.assertEqual(profile_values(["1.1", "2.2", "3.3"]), DataType.FLOAT)
        self.assertEqual(profile_values(["1.1", "2.2", "3.3", ""]), DataType.FLOAT)
        self.assertEqual(
            profile_values(["1", "2.2", "3"]), DataType.FLOAT
        )  # Mixed int and float

        # Test string values
        self.assertEqual(profile_values(["abc", "def"]), DataType.STRING)
        self.assertEqual(profile_values(["abc", "def", ""]), DataType.STRING)

        # Test mixed types
        self.assertEqual(profile_values(["1", "2.2", "abc"]), DataType.MIXED)
        self.assertEqual(profile_values(["1", "2.2", "abc", ""]), DataType.MIXED)

        # Test invalid input
        with self.assertRaises(ValueError):
            profile_values("not iterable")

        # Test with trim=False
        self.assertEqual(
            profile_values(["  true  ", "  false  "], trim=False), DataType.STRING
        )
        self.assertEqual(
            profile_values(["  1  ", "  2  "], trim=False), DataType.STRING
        )


class TestUtilityFunctions(unittest.TestCase):
    """Test cases for utility functions"""

    def test_is_list_like(self):
        """Test list-like detection"""
        # Test list types
        self.assertTrue(is_list_like([]))
        self.assertTrue(is_list_like([1, 2, 3]))

        # Test non-list types
        self.assertFalse(is_list_like({}))
        self.assertFalse(is_list_like("abc"))
        self.assertFalse(is_list_like(None))
        self.assertFalse(is_list_like(123))

    def test_is_dict_like(self):
        """Test dict-like detection"""
        # Test dict types
        self.assertTrue(is_dict_like({}))
        self.assertTrue(is_dict_like({"a": 1}))

        # Test non-dict types
        self.assertFalse(is_dict_like([]))
        self.assertFalse(is_dict_like("abc"))
        self.assertFalse(is_dict_like(None))
        self.assertFalse(is_dict_like(123))

    def test_is_iterable(self):
        """Test iterable detection"""
        # Test iterable types
        self.assertTrue(is_iterable([]))
        self.assertTrue(is_iterable({}))
        self.assertTrue(is_iterable("abc"))
        self.assertTrue(is_iterable((1, 2, 3)))

        # Test non-iterable types
        self.assertFalse(is_iterable(None))
        self.assertFalse(is_iterable(123))
        self.assertFalse(is_iterable(True))

    def test_is_iterable_not_string(self):
        """Test non-string iterable detection"""
        # Test non-string iterables
        self.assertTrue(is_iterable_not_string([]))
        self.assertTrue(is_iterable_not_string({}))
        self.assertTrue(is_iterable_not_string((1, 2, 3)))

        # Test strings and non-iterables
        self.assertFalse(is_iterable_not_string("abc"))
        self.assertFalse(is_iterable_not_string(None))
        self.assertFalse(is_iterable_not_string(123))

    def test_is_empty(self):
        """Test empty value detection"""
        # Test empty values
        self.assertTrue(is_empty(None))
        self.assertTrue(is_empty(""))
        self.assertTrue(is_empty(" "))
        self.assertTrue(is_empty([]))
        self.assertTrue(is_empty({}))
        self.assertTrue(is_empty(()))

        # Test non-empty values
        self.assertFalse(is_empty("abc"))
        self.assertFalse(is_empty([1, 2, 3]))
        self.assertFalse(is_empty({"a": 1}))
        self.assertFalse(is_empty(0))
        self.assertFalse(is_empty(False))


if __name__ == "__main__":
    unittest.main()



"""Unit tests for DSVHelper class."""

import math
import tempfile
import unittest
from pathlib import Path

from splurge_tools.dsv_helper import DsvHelper
from splurge_tools.type_helper import DataType


class TestDSVHelper(unittest.TestCase):
    """Test cases for DSVHelper class."""

    def test_parse_basic(self):
        """Test basic parsing functionality."""
        content = "a,b,c"
        result = DsvHelper.parse(content, ",")
        self.assertEqual(result, ["a", "b", "c"])

    def test_parse_with_bookend(self):
        """Test parsing with text bookends."""
        content = '"a","b","c"'
        result = DsvHelper.parse(content, ",", bookend='"')
        self.assertEqual(result, ["a", "b", "c"])

    def test_parse_with_strip(self):
        """Test parsing with whitespace stripping."""
        content = " a , b , c "
        result = DsvHelper.parse(content, ",", strip=True)
        self.assertEqual(result, ["a", "b", "c"])

    def test_parse_without_strip(self):
        """Test parsing without whitespace stripping."""
        content = " a , b , c "
        result = DsvHelper.parse(content, ",", strip=False)
        self.assertEqual(result, [" a ", " b ", " c "])

    def test_parses_basic(self):
        """Test parsing multiple strings."""
        content = ["a,b,c", "d,e,f"]
        result = DsvHelper.parses(content, ",")
        self.assertEqual(result, [["a", "b", "c"], ["d", "e", "f"]])

    def test_parses_with_bookend(self):
        """Test parsing multiple strings with bookends."""
        content = ['"a","b","c"', '"d","e","f"']
        result = DsvHelper.parses(content, ",", bookend='"')
        self.assertEqual(result, [["a", "b", "c"], ["d", "e", "f"]])

    def test_parse_file(self):
        """Test parsing from file."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write("a,b,c\nd,e,f")
            temp_path = Path(temp_file.name)

        try:
            result = DsvHelper.parse_file(temp_path, ",")
            self.assertEqual(result, [["a", "b", "c"], ["d", "e", "f"]])
        finally:
            temp_path.unlink()

    def test_parse_file_with_bookend(self):
        """Test parsing from file with bookends."""
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write('"a","b","c"\n"d","e","f"')
            temp_path = Path(temp_file.name)

        try:
            result = DsvHelper.parse_file(temp_path, ",", bookend='"')
            self.assertEqual(result, [["a", "b", "c"], ["d", "e", "f"]])
        finally:
            temp_path.unlink()

    def test_invalid_delimiter(self):
        """Test handling of invalid delimiter."""
        with self.assertRaises(ValueError):
            DsvHelper.parse("a,b,c", "")

    def test_invalid_file_path(self):
        """Test handling of invalid file path."""
        with self.assertRaises(FileNotFoundError):
            DsvHelper.parse_file("nonexistent.txt", ",")

    def test_parse_stream_basic(self):
        """Test parse_stream yields correct chunks and parses all rows."""
        total_rows = 2499
        chunk_size = 500
        lines = [f"a{i},b{i},c{i}" for i in range(total_rows)]
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write("\n".join(lines))
            temp_path = Path(temp_file.name)

        try:
            chunks = list(DsvHelper.parse_stream(temp_path, ",", chunk_size=chunk_size))
            # Should yield 5 chunks: 500, 500, 500, 500, 499
            self.assertEqual(len(chunks), math.ceil(total_rows / chunk_size))
            self.assertEqual(len(chunks[0]), chunk_size)
            self.assertEqual(len(chunks[1]), chunk_size)
            self.assertEqual(len(chunks[2]), chunk_size)
            self.assertEqual(len(chunks[3]), chunk_size)
            self.assertEqual(len(chunks[4]), total_rows - 4 * chunk_size)
            # Check content of first and last row
            self.assertEqual(chunks[0][0], ["a0", "b0", "c0"])
            self.assertEqual(chunks[-1][-1], [f"a{total_rows-1}", f"b{total_rows-1}", f"c{total_rows-1}"])
        finally:
            temp_path.unlink()

    def test_parse_stream_with_header_footer(self):
        """Test parse_stream with header and footer skipping."""
        total_rows = 2499
        chunk_size = 500
        header = ["header1,header2,header3"]
        footer = ["footer1,footer2,footer3"]
        data = [f"a{i},b{i},c{i}" for i in range(total_rows)]
        lines = header + data + footer
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write("\n".join(lines))
            temp_path = Path(temp_file.name)

        try:
            chunks = list(DsvHelper.parse_stream(
                temp_path, ",", chunk_size=chunk_size, skip_header_rows=1, skip_footer_rows=1
            ))
            # Should yield 5 chunks: 500, 500, 500, 500, 499
            self.assertEqual(len(chunks), 5)
            self.assertEqual(len(chunks[0]), chunk_size)
            self.assertEqual(len(chunks[1]), chunk_size)
            self.assertEqual(len(chunks[2]), chunk_size)
            self.assertEqual(len(chunks[3]), chunk_size)
            self.assertEqual(len(chunks[4]), total_rows - 4 * chunk_size)
            # Check first and last data row
            self.assertEqual(chunks[0][0], ["a0", "b0", "c0"])
            self.assertEqual(chunks[-1][-1], [f"a{total_rows-1}", f"b{total_rows-1}", f"c{total_rows-1}"])
        finally:
            temp_path.unlink()

    def test_parse_stream_iteration(self):
        """Test iterating over parse_stream yields correct data."""
        total_rows = 2499
        chunk_size = 500
        lines = [f"a{i},b{i},c{i}" for i in range(total_rows)]
        with tempfile.NamedTemporaryFile(mode="w", delete=False) as temp_file:
            temp_file.write("\n".join(lines))
            temp_path = Path(temp_file.name)

        try:
            # Test iteration over parse_stream
            chunk_count = 0
            total_parsed_rows = 0
            expected_chunks = 5  # 2499 rows / 500 chunk_size = 5 chunks
            
            for chunk in DsvHelper.parse_stream(temp_path, ",", chunk_size=chunk_size):
                chunk_count += 1
                total_parsed_rows += len(chunk)
                
                # Verify each chunk has correct structure
                for row in chunk:
                    self.assertEqual(len(row), 3)  # Each row should have 3 columns
                    self.assertTrue(all(isinstance(cell, str) for cell in row))
                
                # Verify chunk sizes (last chunk may be smaller)
                if chunk_count < expected_chunks:
                    self.assertEqual(len(chunk), chunk_size)
                else:
                    self.assertEqual(len(chunk), total_rows - (expected_chunks - 1) * chunk_size)
            
            # Verify total counts
            self.assertEqual(chunk_count, expected_chunks)
            self.assertEqual(total_parsed_rows, total_rows)
            
            # Verify first and last rows across all chunks
            stream = DsvHelper.parse_stream(temp_path, ",", chunk_size=chunk_size)
            first_chunk = next(stream)
            self.assertEqual(first_chunk[0], ["a0", "b0", "c0"])
            
            # Get last chunk
            last_chunk = None
            for chunk in stream:
                last_chunk = chunk
            
            self.assertEqual(last_chunk[-1], [f"a{total_rows-1}", f"b{total_rows-1}", f"c{total_rows-1}"])
            
        finally:
            temp_path.unlink()

    def test_profile_columns_simple(self):
        """
        Test DsvHelper.profile_columns returns correct column names and datatypes for a simple DSV input.
        """
        # Example DSV data: header + two rows
        data = [
            ["Name", "Age", "Active"],
            ["Alice", "30", "true"],
            ["Bob", "25", "false"]
        ]
        expected = [
            {"name": "Name", "datatype": DataType.STRING.name.upper()},
            {"name": "Age", "datatype": DataType.INTEGER.name.upper()},
            {"name": "Active", "datatype": DataType.BOOLEAN.name.upper()}
        ]
        result = DsvHelper.profile_columns(data)
        self.assertEqual(result, expected)

    def test_profile_columns_all_types(self):
        """
        Test DsvHelper.profile_columns detects all supported datatypes.
        """
        # Header row and one value row for each type
        data = [
            [
                "StringCol", "IntCol", "FloatCol", "BoolCol", "DateCol", "TimeCol", "DateTimeCol", "MixedCol", "EmptyCol", "NoneCol"
            ],
            [
                "hello", "42", "3.14", "true", "2024-07-01", "13:45:00", "2024-07-01T13:45:00", "abc", "", "none"
            ],
            [
                "world", "-7", "2.71", "false", "2023-12-31", "23:59:59", "2023-12-31T23:59:59", "123", "", "null"
            ],
            [
                "!@#", "0", "0.0", "TRUE", "2022-01-01", "00:00:00", "2022-01-01T00:00:00", "3.14", "", "None"
            ],
        ]
        expected = [
            {"name": "StringCol", "datatype": DataType.STRING.name.upper()},
            {"name": "IntCol", "datatype": DataType.INTEGER.name.upper()},
            {"name": "FloatCol", "datatype": DataType.FLOAT.name.upper()},
            {"name": "BoolCol", "datatype": DataType.BOOLEAN.name.upper()},
            {"name": "DateCol", "datatype": DataType.DATE.name.upper()},
            {"name": "TimeCol", "datatype": DataType.TIME.name.upper()},
            {"name": "DateTimeCol", "datatype": DataType.DATETIME.name.upper()},
            {"name": "MixedCol", "datatype": DataType.MIXED.name.upper()},
            {"name": "EmptyCol", "datatype": DataType.EMPTY.name.upper()},
            {"name": "NoneCol", "datatype": DataType.NONE.name.upper()},
        ]
        result = DsvHelper.profile_columns(data)
        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



"""
A utility module for working with DSV (Delimited String Values) files.

Copyright (c) 2025 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""

from os import PathLike
from collections import deque
from typing import Optional, Union, Iterator

from splurge_tools.string_tokenizer import StringTokenizer
from splurge_tools.text_file_helper import TextFileHelper
from splurge_tools.tabular_data_model import TabularDataModel


class DsvHelper:
    """
    Utility class for working with DSV (Delimited String Values) files.

    Provides methods to parse DSV content from strings, lists of strings, and files.
    Supports configurable delimiters, text bookends, and whitespace handling options.
    """

    @staticmethod
    def parse(
        content: str,
        delimiter: str,
        *,
        strip: bool = True,
        bookend: Optional[str] = None,
        bookend_strip: bool = True
    ) -> list[str]:
        """
        Parse a string into a list of strings.

        Args:
            content (str): The string to parse.
            delimiter (str): The delimiter to use.
            strip (bool): Whether to strip whitespace from the strings.
            bookend (Optional[str]): The bookend to use for text fields.
            bookend_strip (bool): Whether to strip whitespace from the bookend.

        Returns:
            list[str]: The list of strings.

        Raises:
            ValueError: If delimiter is empty or None.

        Example:
            >>> DsvHelper.parse("a,b,c", ",")
            ['a', 'b', 'c']
            >>> DsvHelper.parse('"a","b","c"', ",", bookend='"')
            ['a', 'b', 'c']
        """
        if not delimiter:
            raise ValueError("Delimiter must not be empty or None.")

        tokens: list[str] = StringTokenizer.parse(content, delimiter, strip=strip)

        if bookend:
            tokens = [
                StringTokenizer.remove_bookends(token, bookend, strip=bookend_strip)
                for token in tokens
            ]

        return tokens

    @classmethod
    def parses(
        cls,
        content: list[str],
        delimiter: str,
        *,
        strip: bool = True,
        bookend: Optional[str] = None,
        bookend_strip: bool = True
    ) -> list[list[str]]:
        """
        Parse a list of strings into a list of lists of strings.

        Args:
            content (list[str]): The list of strings to parse.
            delimiter (str): The delimiter to use.
            strip (bool): Whether to strip whitespace from the strings.
            bookend (Optional[str]): The bookend to use for text fields.
            bookend_strip (bool): Whether to strip whitespace from the bookend.

        Returns:
            list[list[str]]: The list of lists of strings.

        Raises:
            ValueError: If delimiter is empty or None.
            TypeError: If content is not a list of strings.

        Example:
            >>> DsvHelper.parses(["a,b,c", "d,e,f"], ",")
            [['a', 'b', 'c'], ['d', 'e', 'f']]
        """
        if not isinstance(content, list) or not all(isinstance(item, str) for item in content):
            raise TypeError("Content must be a list of strings.")

        return [
            cls.parse(item, delimiter, strip=strip, bookend=bookend, bookend_strip=bookend_strip)
            for item in content
        ]

    @classmethod
    def parse_file(
        cls,
        file_path: Union[PathLike, str],
        delimiter: str,
        *,
        strip: bool = True,
        bookend: Optional[str] = None,
        bookend_strip: bool = True,
        encoding: str = "utf-8",
        skip_header_rows: int = 0,
        skip_footer_rows: int = 0
    ) -> list[list[str]]:
        """
        Parse a file into a list of lists of strings.

        Args:
            file_path (Union[PathLike, str]): The path to the file to parse.
            delimiter (str): The delimiter to use.
            strip (bool): Whether to strip whitespace from the strings.
            bookend (Optional[str]): The bookend to use for text fields.
            bookend_strip (bool): Whether to strip whitespace from the bookend.
            encoding (str): The file encoding.
            skip_header_rows (int): Number of header rows to skip.
            skip_footer_rows (int): Number of footer rows to skip.

        Returns:
            list[list[str]]: The list of lists of strings.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If delimiter is empty or None.
            PermissionError: If the file cannot be accessed.

        Example:
            >>> DsvHelper.parse_file("data.csv", ",")
            [['header1', 'header2'], ['value1', 'value2']]
        """
        lines: list[str] = TextFileHelper.load(
            file_path,
            encoding=encoding,
            skip_header_rows=skip_header_rows,
            skip_footer_rows=skip_footer_rows
        )

        return cls.parses(
            lines,
            delimiter,
            strip=strip,
            bookend=bookend,
            bookend_strip=bookend_strip
        )

    @classmethod
    def parse_stream(
        cls,
        file_path: Union[PathLike, str],
        delimiter: str,
        *,
        strip: bool = True,
        bookend: Optional[str] = None,
        bookend_strip: bool = True,
        encoding: str = "utf-8",
        skip_header_rows: int = 0,
        skip_footer_rows: int = 0,
        chunk_size: int = 100
    ) -> Iterator[list[list[str]]]:
        """
        Stream-parse a DSV file in chunks of lines.

        Args:
            file_path (Union[PathLike, str]): The path to the file to parse.
            delimiter (str): The delimiter to use.
            strip (bool): Whether to strip whitespace from the strings.
            bookend (Optional[str]): The bookend to use for text fields.
            bookend_strip (bool): Whether to strip whitespace from the bookend.
            encoding (str): The file encoding.
            skip_header_rows (int): Number of header rows to skip.
            skip_footer_rows (int): Number of footer rows to skip.
            chunk_size (int): Number of lines per chunk (default: 100).

        Yields:
            list[list[str]]: Parsed rows for each chunk.

        Raises:
            FileNotFoundError: If the file does not exist.
            ValueError: If delimiter is empty or None.
            PermissionError: If the file cannot be accessed.
        """
        if not delimiter:
            raise ValueError("Delimiter must not be empty or None.")
        if chunk_size < 100:
            raise ValueError("chunk_size must be at least 100.")
        if skip_header_rows < 0 or skip_footer_rows < 0:
            raise ValueError("skip_header_rows and skip_footer_rows must be >= 0.")

        with open(file_path, "r", encoding=encoding) as stream:
            # Skip header rows
            for _ in range(skip_header_rows):
                if not stream.readline():
                    return

            # If skipping footer rows, use a buffer
            if skip_footer_rows > 0:
                buffer = deque()  
                chunk = []  
                for line in stream:
                    buffer.append(line.strip() if strip else line.rstrip("\n"))
                    if len(buffer) > skip_footer_rows:
                        chunk.append(buffer.popleft())
                        if len(chunk) == chunk_size:
                            yield cls.parses(
                                chunk,
                                delimiter,
                                strip=strip,
                                bookend=bookend,
                                bookend_strip=bookend_strip
                            )
                            chunk = []
                # Yield any remaining chunk (excluding footer rows)
                if chunk:
                    yield cls.parses(
                        chunk,
                        delimiter,
                        strip=strip,
                        bookend=bookend,
                        bookend_strip=bookend_strip
                    )
            else:
                chunk = []
                for line in stream:
                    chunk.append(line.strip() if strip else line.rstrip("\n"))
                    if len(chunk) == chunk_size:
                        yield cls.parses(
                            chunk,
                            delimiter,
                            strip=strip,
                            bookend=bookend,
                            bookend_strip=bookend_strip
                        )
                        chunk = []
                if chunk:
                    yield cls.parses(
                        chunk,
                        delimiter,
                        strip=strip,
                        bookend=bookend,
                        bookend_strip=bookend_strip
                    )

    @classmethod
    def profile_columns(
        cls,
        data: list[list[str]],
        *,
        header_rows: int = 1,
        multi_row_headers: int = 1,
        skip_empty_rows: bool = True
    ) -> list[dict[str, str]]:
        """
        Generate a simple data profile from parsed DSV data.

        Args:
            data (list[list[str]]): Parsed DSV data (including header row).
            header_rows (int): Number of header rows (default: 1).
            multi_row_headers (int): Number of rows to merge for column names (default: 1).
            skip_empty_rows (bool): Whether to skip empty rows (default: True).

        Returns:
            list[dict[str, str]]: List of dicts with column name and inferred datatype.
                Example: [{"name": "Column_1", "datatype": "STRING"}, ...]
        """
        model = TabularDataModel(
            data,
            header_rows=header_rows,
            multi_row_headers=multi_row_headers,
            skip_empty_rows=skip_empty_rows
        )

        profile: list[dict[str, str]] = []
        for col in model.column_names:
            dtype = model.column_type(col)
            profile.append({"name": col, "datatype": dtype.name.upper()})
        return profile



"""
case_helper.py

A utility module for case conversion operations.

Copyright (c) 2025 Jim Schilling

Please preserve this header and all related material when sharing!

This software is licensed under the MIT License.
"""

from functools import wraps
from typing import Callable, Any


def handle_empty_value(
    func: Callable[[str, Any], str]
) -> Callable[[str, Any], str]:
    """
    Decorator to handle empty value checks for case conversion methods.
    Returns empty string if input value is None or empty.
    """
    @wraps(func)
    def wrapper(
        value: str,
        *args: Any,
        **kwargs: Any
    ) -> str:
        if value is None or not value:
            return ""
        return func(value, *args, **kwargs)

    return wrapper


class CaseHelper:
    """
    A utility class for case conversion operations.

    This class provides methods to:
    - Convert strings to different cases (train, sentence, camel, snake, kebab, pascal)
    - Handle case-insensitive comparisons
    - Normalize string formatting by converting separators to spaces

    All methods support an optional normalize parameter (default: True) that:
    - Converts underscores and hyphens to spaces before processing
    - Ensures consistent handling of mixed input formats
    """

    @staticmethod
    def normalize(
        value: str
    ) -> str:
        """
        Normalize a string by converting underscores and hyphens to spaces.
        This ensures consistent handling of mixed input formats.

        Args:
            value: Input string that may contain underscores or hyphens

        Returns:
            String with underscores and hyphens converted to spaces

        Example:
            "hello_world" -> "hello world"
            "hello-world" -> "hello world"
        """
        return value.replace("_", " ").replace("-", " ")

    @classmethod
    @handle_empty_value
    def to_train(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to train case (capitalized words separated by hyphens).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in train case format

        Example:
            "hello world" -> "Hello-World"
            "hello_world" -> "Hello-World"
        """
        if normalize:
            value = cls.normalize(value)
        return value.title().replace(" ", "-")

    @classmethod
    @handle_empty_value
    def to_sentence(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to sentence case (first word capitalized, rest lowercase).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in sentence case format

        Example:
            "hello world" -> "Hello world"
            "hello_world" -> "Hello world"
        """
        if normalize:
            value = cls.normalize(value)
        return value.capitalize()

    @classmethod
    @handle_empty_value
    def to_camel(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to camel case (first word lowercase, subsequent words capitalized).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in camel case format

        Example:
            "hello world" -> "helloWorld"
            "hello_world" -> "helloWorld"
        """
        if normalize:
            value = cls.normalize(value)

        words: list[str] = value.split()
        if not words:
            return ""
        
        return words[0].lower() + "".join(word.title() for word in words[1:])

    @classmethod
    @handle_empty_value
    def to_snake(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to snake case (all lowercase with underscore separators).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in snake case format

        Example:
            "hello world" -> "hello_world"
            "hello-world" -> "hello_world"
        """
        if normalize:
            value = cls.normalize(value)            
        return value.lower().replace(" ", "_")

    @classmethod
    @handle_empty_value
    def to_kebab(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to kebab case (all lowercase with hyphen separators).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in kebab case format

        Example:
            "hello world" -> "hello-world"
            "hello_world" -> "hello-world"
        """
        if normalize:
            value = cls.normalize(value)
        return value.lower().replace(" ", "-")

    @classmethod
    @handle_empty_value
    def to_pascal(
        cls,
        value: str,
        *,
        normalize: bool = True
    ) -> str:
        """
        Convert a string to pascal case (all words capitalized, no separators).

        Args:
            value: Input string to convert
            normalize: If True, converts underscores and hyphens to spaces first

        Returns:
            String in pascal case format

        Example:
            "hello world" -> "HelloWorld"
            "hello_world" -> "HelloWorld"
        """
        if normalize:
            value = cls.normalize(value)
        return "".join(word.title() for word in value.split())
